"""
    pygments.lexers.bdd
    ~~~~~~~~~~~~~~~~~~~

    Lexer for BDD(Behavior-driven development).

    :copyright: Copyright 2006-2024 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.lexer import RegexLexer, include
from pygments.token import Comment, Keyword, Name, String, Number, Text, \
    Punctuation, Whitespace

__all__ = ['BddLexer']

class BddLexer(RegexLexer):
    """
    Lexer for BDD(Behavior-driven development), which highlights not only
    keywords, but also comments, punctuations, strings, numbers, and variables.
    """

    name = 'Bdd'
    aliases = ['bdd']
    filenames = ['*.feature']
    mimetypes = ['text/x-bdd']
    url = 'https://en.wikipedia.org/wiki/Behavior-driven_development'
    version_added = '2.11'

    step_keywords = (r'Given|When|Then|Add|And|Feature|Scenario Outline|'
                     r'Scenario|Background|Examples|But')

    tokens = {
        'comments': [
            (r'^\s*#.*$', Comment),
        ],
        'miscellaneous': [
            (r'(<|>|\[|\]|=|\||:|\(|\)|\{|\}|,|\.|;|-|_|\$)', Punctuation),
            (r'((?<=\<)[^\\>]+(?=\>))', Name.Variable),
            (r'"([^\"]*)"', String),
            (r'^@\S+', Name.Label),
        ],
        'numbers': [
            (r'(\d+\.?\d*|\d*\.\d+)([eE][+-]?[0-9]+)?', Number),
        ],
        'root': [
            (r'\n|\s+', Whitespace),
            (step_keywords, Keyword),
            include('comments'),
            include('miscellaneous'),
            include('numbers'),
            (r'\S+', Text),
        ]
    }

    def analyse_text(self, text):
        return



"""
    pygments.lexers.testing
    ~~~~~~~~~~~~~~~~~~~~~~~

    Lexers for testing languages.

    :copyright: Copyright 2006-2024 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.lexer import RegexLexer, include, bygroups
from pygments.token import Comment, Keyword, Name, String, Number, Generic, Text

__all__ = ['GherkinLexer', 'TAPLexer']


class GherkinLexer(RegexLexer):
    """
    For Gherkin syntax.
    """
    name = 'Gherkin'
    aliases = ['gherkin', 'cucumber']
    filenames = ['*.feature']
    mimetypes = ['text/x-gherkin']
    url = 'https://cucumber.io/docs/gherkin'
    version_added = '1.2'

    feature_keywords = '^(기능|機能|功能|フィーチャ|خاصية|תכונה|Функціонал|Функционалност|Функционал|Фича|Особина|Могућност|Özellik|Właściwość|Tính năng|Trajto|Savybė|Požiadavka|Požadavek|Osobina|Ominaisuus|Omadus|OH HAI|Mogućnost|Mogucnost|Jellemző|Fīča|Funzionalità|Funktionalität|Funkcionalnost|Funkcionalitāte|Funcționalitate|Functionaliteit|Functionalitate|Funcionalitat|Funcionalidade|Fonctionnalité|Fitur|Feature|Egenskap|Egenskab|Crikey|Característica|Arwedd)(:)(.*)$'
    feature_element_keywords = '^(\\s*)(시나리오 개요|시나리오|배경|背景|場景大綱|場景|场景大纲|场景|劇本大綱|劇本|剧本大纲|剧本|テンプレ|シナリオテンプレート|シナリオテンプレ|シナリオアウトライン|シナリオ|سيناريو مخطط|سيناريو|الخلفية|תרחיש|תבנית תרחיש|רקע|Тарих|Сценарій|Сценарио|Сценарий структураси|Сценарий|Структура сценарію|Структура сценарија|Структура сценария|Скица|Рамка на сценарий|Пример|Предыстория|Предистория|Позадина|Передумова|Основа|Концепт|Контекст|Założenia|Wharrimean is|Tình huống|The thing of it is|Tausta|Taust|Tapausaihio|Tapaus|Szenariogrundriss|Szenario|Szablon scenariusza|Stsenaarium|Struktura scenarija|Skica|Skenario konsep|Skenario|Situācija|Senaryo taslağı|Senaryo|Scénář|Scénario|Schema dello scenario|Scenārijs pēc parauga|Scenārijs|Scenár|Scenaro|Scenariusz|Scenariul de şablon|Scenariul de sablon|Scenariu|Scenario Outline|Scenario Amlinellol|Scenario|Scenarijus|Scenarijaus šablonas|Scenarij|Scenarie|Rerefons|Raamstsenaarium|Primer|Pozadí|Pozadina|Pozadie|Plan du scénario|Plan du Scénario|Osnova scénáře|Osnova|Náčrt Scénáře|Náčrt Scenáru|Mate|MISHUN SRSLY|MISHUN|Kịch bản|Konturo de la scenaro|Kontext|Konteksts|Kontekstas|Kontekst|Koncept|Khung tình huống|Khung kịch bản|Háttér|Grundlage|Geçmiş|Forgatókönyv vázlat|Forgatókönyv|Fono|Esquema do Cenário|Esquema do Cenario|Esquema del escenario|Esquema de l\'escenari|Escenario|Escenari|Dis is what went down|Dasar|Contexto|Contexte|Contesto|Condiţii|Conditii|Cenário|Cenario|Cefndir|Bối cảnh|Blokes|Bakgrunn|Bakgrund|Baggrund|Background|B4|Antecedents|Antecedentes|All y\'all|Achtergrond|Abstrakt Scenario|Abstract Scenario)(:)(.*)$'
    examples_keywords = '^(\\s*)(예|例子|例|サンプル|امثلة|דוגמאות|Сценарији|Примери|Приклади|Мисоллар|Значения|Örnekler|Voorbeelden|Variantai|Tapaukset|Scenarios|Scenariji|Scenarijai|Příklady|Példák|Príklady|Przykłady|Primjeri|Primeri|Piemēri|Pavyzdžiai|Paraugs|Juhtumid|Exemplos|Exemples|Exemplele|Exempel|Examples|Esempi|Enghreifftiau|Ekzemploj|Eksempler|Ejemplos|EXAMPLZ|Dữ liệu|Contoh|Cobber|Beispiele)(:)(.*)$'
    step_keywords = '^(\\s*)(하지만|조건|먼저|만일|만약|단|그리고|그러면|那麼|那么|而且|當|当|前提|假設|假设|假如|假定|但是|但し|並且|并且|同時|同时|もし|ならば|ただし|しかし|かつ|و |متى |لكن |عندما |ثم |بفرض |اذاً |כאשר |וגם |בהינתן |אזי |אז |אבל |Якщо |Унда |То |Припустимо, що |Припустимо |Онда |Но |Нехай |Лекин |Когато |Када |Кад |К тому же |И |Задато |Задати |Задате |Если |Допустим |Дадено |Ва |Бирок |Аммо |Али |Але |Агар |А |І |Și |És |Zatati |Zakładając |Zadato |Zadate |Zadano |Zadani |Zadan |Youse know when youse got |Youse know like when |Yna |Ya know how |Ya gotta |Y |Wun |Wtedy |When y\'all |When |Wenn |WEN |Và |Ve |Und |Un |Thì |Then y\'all |Then |Tapi |Tak |Tada |Tad |Så |Stel |Soit |Siis |Si |Sed |Se |Quando |Quand |Quan |Pryd |Pokud |Pokiaľ |Però |Pero |Pak |Oraz |Onda |Ond |Oletetaan |Og |Och |O zaman |Når |När |Niin |Nhưng |N |Mutta |Men |Mas |Maka |Majd |Mais |Maar |Ma |Lorsque |Lorsqu\'|Kun |Kuid |Kui |Khi |Keď |Ketika |Když |Kaj |Kai |Kada |Kad |Jeżeli |Ja |Ir |I CAN HAZ |I |Ha |Givun |Givet |Given y\'all |Given |Gitt |Gegeven |Gegeben sei |Fakat |Eğer ki |Etant donné |Et |Então |Entonces |Entao |En |Eeldades |E |Duota |Dun |Donitaĵo |Donat |Donada |Do |Diyelim ki |Dengan |Den youse gotta |De |Dato |Dar |Dann |Dan |Dado |Dacă |Daca |DEN |Când |Cuando |Cho |Cept |Cand |Cal |But y\'all |But |Buh |Biết |Bet |BUT |Atès |Atunci |Atesa |Anrhegedig a |Angenommen |And y\'all |And |An |Ama |Als |Alors |Allora |Ali |Aleshores |Ale |Akkor |Aber |AN |A také |A |\\* )'

    tokens = {
        'comments': [
            (r'^\s*#.*$', Comment),
        ],
        'feature_elements': [
            (step_keywords, Keyword, "step_content_stack"),
            include('comments'),
            (r"(\s|.)", Name.Function),
        ],
        'feature_elements_on_stack': [
            (step_keywords, Keyword, "#pop:2"),
            include('comments'),
            (r"(\s|.)", Name.Function),
        ],
        'examples_table': [
            (r"\s+\|", Keyword, 'examples_table_header'),
            include('comments'),
            (r"(\s|.)", Name.Function),
        ],
        'examples_table_header': [
            (r"\s+\|\s*$", Keyword, "#pop:2"),
            include('comments'),
            (r"\\\|", Name.Variable),
            (r"\s*\|", Keyword),
            (r"[^|]", Name.Variable),
        ],
        'scenario_sections_on_stack': [
            (feature_element_keywords,
             bygroups(Name.Function, Keyword, Keyword, Name.Function),
             "feature_elements_on_stack"),
        ],
        'narrative': [
            include('scenario_sections_on_stack'),
            include('comments'),
            (r"(\s|.)", Name.Function),
        ],
        'table_vars': [
            (r'(<[^>]+>)', Name.Variable),
        ],
        'numbers': [
            (r'(\d+\.?\d*|\d*\.\d+)([eE][+-]?[0-9]+)?', String),
        ],
        'string': [
            include('table_vars'),
            (r'(\s|.)', String),
        ],
        'py_string': [
            (r'"""', Keyword, "#pop"),
            include('string'),
        ],
        'step_content_root': [
            (r"$", Keyword, "#pop"),
            include('step_content'),
        ],
        'step_content_stack': [
            (r"$", Keyword, "#pop:2"),
            include('step_content'),
        ],
        'step_content': [
            (r'"', Name.Function, "double_string"),
            include('table_vars'),
            include('numbers'),
            include('comments'),
            (r'(\s|.)', Name.Function),
        ],
        'table_content': [
            (r"\s+\|\s*$", Keyword, "#pop"),
            include('comments'),
            (r"\\\|", String),
            (r"\s*\|", Keyword),
            include('string'),
        ],
        'double_string': [
            (r'"', Name.Function, "#pop"),
            include('string'),
        ],
        'root': [
            (r'\n', Name.Function),
            include('comments'),
            (r'"""', Keyword, "py_string"),
            (r'\s+\|', Keyword, 'table_content'),
            (r'"', Name.Function, "double_string"),
            include('table_vars'),
            include('numbers'),
            (r'(\s*)(@[^@\r\n\t ]+)', bygroups(Name.Function, Name.Tag)),
            (step_keywords, bygroups(Name.Function, Keyword),
             'step_content_root'),
            (feature_keywords, bygroups(Keyword, Keyword, Name.Function),
             'narrative'),
            (feature_element_keywords,
             bygroups(Name.Function, Keyword, Keyword, Name.Function),
             'feature_elements'),
            (examples_keywords,
             bygroups(Name.Function, Keyword, Keyword, Name.Function),
             'examples_table'),
            (r'(\s|.)', Name.Function),
        ]
    }

    def analyse_text(self, text):
        return


class TAPLexer(RegexLexer):
    """
    For Test Anything Protocol (TAP) output.
    """
    name = 'TAP'
    url = 'https://testanything.org/'
    aliases = ['tap']
    filenames = ['*.tap']
    version_added = '2.1'

    tokens = {
        'root': [
            # A TAP version may be specified.
            (r'^TAP version \d+\n', Name.Namespace),

            # Specify a plan with a plan line.
            (r'^1\.\.\d+', Keyword.Declaration, 'plan'),

            # A test failure
            (r'^(not ok)([^\S\n]*)(\d*)',
             bygroups(Generic.Error, Text, Number.Integer), 'test'),

            # A test success
            (r'^(ok)([^\S\n]*)(\d*)',
             bygroups(Keyword.Reserved, Text, Number.Integer), 'test'),

            # Diagnostics start with a hash.
            (r'^#.*\n', Comment),

            # TAP's version of an abort statement.
            (r'^Bail out!.*\n', Generic.Error),

            # TAP ignores any unrecognized lines.
            (r'^.*\n', Text),
        ],
        'plan': [
            # Consume whitespace (but not newline).
            (r'[^\S\n]+', Text),

            # A plan may have a directive with it.
            (r'#', Comment, 'directive'),

            # Or it could just end.
            (r'\n', Comment, '#pop'),

            # Anything else is wrong.
            (r'.*\n', Generic.Error, '#pop'),
        ],
        'test': [
            # Consume whitespace (but not newline).
            (r'[^\S\n]+', Text),

            # A test may have a directive with it.
            (r'#', Comment, 'directive'),

            (r'\S+', Text),

            (r'\n', Text, '#pop'),
        ],
        'directive': [
            # Consume whitespace (but not newline).
            (r'[^\S\n]+', Comment),

            # Extract todo items.
            (r'(?i)\bTODO\b', Comment.Preproc),

            # Extract skip items.
            (r'(?i)\bSKIP\S*', Comment.Preproc),

            (r'\S+', Comment),

            (r'\n', Comment, '#pop:2'),
        ],
    }



"""
Backward-compatability shims to support Python 3.9 and earlier.
"""

from jaraco.test.cpython import from_test_support, try_import

import_helper = try_import('import_helper') or from_test_support(
    'modules_setup', 'modules_cleanup', 'DirsOnSysPath'
)
os_helper = try_import('os_helper') or from_test_support('temp_dir')
warnings_helper = try_import('warnings_helper') or from_test_support(
    'ignore_warnings', 'check_warnings'
)



# $Id: writer_aux.py 9037 2022-03-05 23:31:10Z milde $
# Author: Lea Wiemann <LeWiemann@gmail.com>
# Copyright: This module has been placed in the public domain.

"""
Auxiliary transforms mainly to be used by Writer components.

This module is called "writer_aux" because otherwise there would be
conflicting imports like this one::

    from docutils import writers
    from docutils.transforms import writers
"""

__docformat__ = 'reStructuredText'

import warnings

from docutils import nodes, languages
from docutils.transforms import Transform


class Compound(Transform):

    """
    .. warning:: This transform is not used by Docutils since Dec 2010
                 and will be removed in Docutils 0.21 or later.

    Flatten all compound paragraphs.  For example, transform ::

        <compound>
            <paragraph>
            <literal_block>
            <paragraph>

    into ::

        <paragraph>
        <literal_block classes="continued">
        <paragraph classes="continued">
    """

    default_priority = 910

    def __init__(self, document, startnode=None):
        warnings.warn('docutils.transforms.writer_aux.Compound is deprecated'
                      ' and will be removed in Docutils 0.21 or later.',
                      DeprecationWarning, stacklevel=2)
        super().__init__(document, startnode)

    def apply(self):
        for compound in self.document.findall(nodes.compound):
            first_child = True
            for child in compound:
                if first_child:
                    if not isinstance(child, nodes.Invisible):
                        first_child = False
                else:
                    child['classes'].append('continued')
            # Substitute children for compound.
            compound.replace_self(compound[:])


class Admonitions(Transform):

    """
    Transform specific admonitions, like this:

        <note>
            <paragraph>
                 Note contents ...

    into generic admonitions, like this::

        <admonition classes="note">
            <title>
                Note
            <paragraph>
                Note contents ...

    The admonition title is localized.
    """

    default_priority = 920

    def apply(self):
        language = languages.get_language(self.document.settings.language_code,
                                          self.document.reporter)
        for node in self.document.findall(nodes.Admonition):
            node_name = node.__class__.__name__
            # Set class, so that we know what node this admonition came from.
            node['classes'].append(node_name)
            if not isinstance(node, nodes.admonition):
                # Specific admonition.  Transform into a generic admonition.
                admonition = nodes.admonition(node.rawsource, *node.children,
                                              **node.attributes)
                title = nodes.title('', language.labels[node_name])
                admonition.insert(0, title)
                node.replace_self(admonition)



import email.message
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generator
from typing import List
from typing import Sequence
from typing import TextIO
from typing import Tuple

def must_decode(value: bytes | str) -> str: ...
def parse(fp: TextIO) -> email.message.Message: ...
def get(msg: email.message.Message, header: str) -> str: ...
def get_all(msg: email.message.Message, header: str) -> List[str]: ...

_header_attr_triple = Tuple[str, str, bool]
_header_attrs = Sequence[_header_attr_triple]
HEADER_ATTRS_1_0: _header_attrs
HEADER_ATTRS_1_1: _header_attrs
HEADER_ATTRS_1_2: _header_attrs
HEADER_ATTRS_2_0 = HEADER_ATTRS_1_2
HEADER_ATTRS_2_1: _header_attrs
HEADER_ATTRS_2_2: _header_attrs
HEADER_ATTRS_2_3: _header_attrs
HEADER_ATTRS: Dict[str, _header_attrs]
_metadata_version = Tuple[int, int]
METADATA_VERSIONS: Sequence[_metadata_version]
MAX_METADATA_VERSION: _metadata_version
MAX_METADATA_VERSION_STR: str

class UnknownMetadataVersion(UserWarning):
    metadata_version: str
    def __init__(self, metadata_version: str) -> None: ...

class NewMetadataVersion(UserWarning):
    metadata_version: str
    def __init__(self, metadata_version: str) -> None: ...

class Distribution:
    metadata_version: str | None
    name: str | None
    version: str | None
    platforms: Sequence[str]
    supported_platforms: Sequence[str]
    summary: str | None
    description: str | None
    keywords: str | None
    home_page: str | None
    download_url: str | None
    author: str | None
    author_email: str | None
    license: str | None
    classifiers: Sequence[str]
    requires: Sequence[str]
    provides: Sequence[str]
    obsoletes: Sequence[str]
    maintainer: str | None
    maintainer_email: str | None
    requires_python: str | None
    requires_external: Sequence[str]
    requires_dist: Sequence[str]
    provides_dist: Sequence[str]
    obsoletes_dist: Sequence[str]
    project_urls: Sequence[str]
    provides_extras: Sequence[str]
    description_content_type: str | None
    dynamic: Sequence[str]
    def extractMetadata(self) -> None: ...
    def read(self) -> bytes: ...
    def parse(self, data: bytes) -> None: ...
    def __iter__(self) -> Generator[str, None, None]: ...
    def iterkeys(self) -> Generator[str, None, None]: ...



import pathlib
import tempfile

import pytest

@pytest.fixture(scope="module")
def here():
    file_path = pathlib.Path(__file__)
    return file_path.parent

@pytest.fixture(scope="module")
def examples_dir(here):
    return here / "../../docs/examples"

_ARCHIVE_FILENAMES = [
    'mypackage-0.1.tar',
    'mypackage-0.1.tar.gz',
    'mypackage-0.1.tar.bz2',
    'mypackage-0.1.zip',
]
@pytest.fixture(scope="module", params=_ARCHIVE_FILENAMES)
def archive(examples_dir, request):
    yield examples_dir / request.param

@pytest.fixture(scope="module")
def test_egg(examples_dir):
    yield examples_dir / 'mypackage-0.1-py2.6.egg'

@pytest.fixture(scope="module")
def test_wheel(examples_dir):
    yield examples_dir / 'mypackage-0.1-cp26-none-linux_x86_64.whl'

@pytest.fixture()
def temp_dir():
    with tempfile.TemporaryDirectory() as tmpdir:
        return pathlib.Path(tmpdir)




from enum import IntEnum
from functools import lru_cache
from itertools import filterfalse
from logging import getLogger
from operator import attrgetter
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

from .cells import (
    _is_single_cell_widths,
    cached_cell_len,
    cell_len,
    get_character_cell_size,
    set_cell_size,
)
from .repr import Result, rich_repr
from .style import Style

if TYPE_CHECKING:
    from .console import Console, ConsoleOptions, RenderResult

log = getLogger("rich")


class ControlType(IntEnum):
    """Non-printable control codes which typically translate to ANSI codes."""

    BELL = 1
    CARRIAGE_RETURN = 2
    HOME = 3
    CLEAR = 4
    SHOW_CURSOR = 5
    HIDE_CURSOR = 6
    ENABLE_ALT_SCREEN = 7
    DISABLE_ALT_SCREEN = 8
    CURSOR_UP = 9
    CURSOR_DOWN = 10
    CURSOR_FORWARD = 11
    CURSOR_BACKWARD = 12
    CURSOR_MOVE_TO_COLUMN = 13
    CURSOR_MOVE_TO = 14
    ERASE_IN_LINE = 15
    SET_WINDOW_TITLE = 16


ControlCode = Union[
    Tuple[ControlType],
    Tuple[ControlType, Union[int, str]],
    Tuple[ControlType, int, int],
]


@rich_repr()
class Segment(NamedTuple):
    """A piece of text with associated style. Segments are produced by the Console render process and
    are ultimately converted in to strings to be written to the terminal.

    Args:
        text (str): A piece of text.
        style (:class:`~rich.style.Style`, optional): An optional style to apply to the text.
        control (Tuple[ControlCode], optional): Optional sequence of control codes.

    Attributes:
        cell_length (int): The cell length of this Segment.
    """

    text: str
    style: Optional[Style] = None
    control: Optional[Sequence[ControlCode]] = None

    @property
    def cell_length(self) -> int:
        """The number of terminal cells required to display self.text.

        Returns:
            int: A number of cells.
        """
        text, _style, control = self
        return 0 if control else cell_len(text)

    def __rich_repr__(self) -> Result:
        yield self.text
        if self.control is None:
            if self.style is not None:
                yield self.style
        else:
            yield self.style
            yield self.control

    def __bool__(self) -> bool:
        """Check if the segment contains text."""
        return bool(self.text)

    @property
    def is_control(self) -> bool:
        """Check if the segment contains control codes."""
        return self.control is not None

    @classmethod
    @lru_cache(1024 * 16)
    def _split_cells(cls, segment: "Segment", cut: int) -> Tuple["Segment", "Segment"]:
        """Split a segment in to two at a given cell position.

        Note that splitting a double-width character, may result in that character turning
        into two spaces.

        Args:
            segment (Segment): A segment to split.
            cut (int): A cell position to cut on.

        Returns:
            A tuple of two segments.
        """
        text, style, control = segment
        _Segment = Segment
        cell_length = segment.cell_length
        if cut >= cell_length:
            return segment, _Segment("", style, control)

        cell_size = get_character_cell_size

        pos = int((cut / cell_length) * len(text))

        while True:
            before = text[:pos]
            cell_pos = cell_len(before)
            out_by = cell_pos - cut
            if not out_by:
                return (
                    _Segment(before, style, control),
                    _Segment(text[pos:], style, control),
                )
            if out_by == -1 and cell_size(text[pos]) == 2:
                return (
                    _Segment(text[:pos] + " ", style, control),
                    _Segment(" " + text[pos + 1 :], style, control),
                )
            if out_by == +1 and cell_size(text[pos - 1]) == 2:
                return (
                    _Segment(text[: pos - 1] + " ", style, control),
                    _Segment(" " + text[pos:], style, control),
                )
            if cell_pos < cut:
                pos += 1
            else:
                pos -= 1

    def split_cells(self, cut: int) -> Tuple["Segment", "Segment"]:
        """Split segment in to two segments at the specified column.

        If the cut point falls in the middle of a 2-cell wide character then it is replaced
        by two spaces, to preserve the display width of the parent segment.

        Returns:
            Tuple[Segment, Segment]: Two segments.
        """
        text, style, control = self

        if _is_single_cell_widths(text):
            # Fast path with all 1 cell characters
            if cut >= len(text):
                return self, Segment("", style, control)
            return (
                Segment(text[:cut], style, control),
                Segment(text[cut:], style, control),
            )

        return self._split_cells(self, cut)

    @classmethod
    def line(cls) -> "Segment":
        """Make a new line segment."""
        return cls("\n")

    @classmethod
    def apply_style(
        cls,
        segments: Iterable["Segment"],
        style: Optional[Style] = None,
        post_style: Optional[Style] = None,
    ) -> Iterable["Segment"]:
        """Apply style(s) to an iterable of segments.

        Returns an iterable of segments where the style is replaced by ``style + segment.style + post_style``.

        Args:
            segments (Iterable[Segment]): Segments to process.
            style (Style, optional): Base style. Defaults to None.
            post_style (Style, optional): Style to apply on top of segment style. Defaults to None.

        Returns:
            Iterable[Segments]: A new iterable of segments (possibly the same iterable).
        """
        result_segments = segments
        if style:
            apply = style.__add__
            result_segments = (
                cls(text, None if control else apply(_style), control)
                for text, _style, control in result_segments
            )
        if post_style:
            result_segments = (
                cls(
                    text,
                    (
                        None
                        if control
                        else (_style + post_style if _style else post_style)
                    ),
                    control,
                )
                for text, _style, control in result_segments
            )
        return result_segments

    @classmethod
    def filter_control(
        cls, segments: Iterable["Segment"], is_control: bool = False
    ) -> Iterable["Segment"]:
        """Filter segments by ``is_control`` attribute.

        Args:
            segments (Iterable[Segment]): An iterable of Segment instances.
            is_control (bool, optional): is_control flag to match in search.

        Returns:
            Iterable[Segment]: And iterable of Segment instances.

        """
        if is_control:
            return filter(attrgetter("control"), segments)
        else:
            return filterfalse(attrgetter("control"), segments)

    @classmethod
    def split_lines(cls, segments: Iterable["Segment"]) -> Iterable[List["Segment"]]:
        """Split a sequence of segments in to a list of lines.

        Args:
            segments (Iterable[Segment]): Segments potentially containing line feeds.

        Yields:
            Iterable[List[Segment]]: Iterable of segment lists, one per line.
        """
        line: List[Segment] = []
        append = line.append

        for segment in segments:
            if "\n" in segment.text and not segment.control:
                text, style, _ = segment
                while text:
                    _text, new_line, text = text.partition("\n")
                    if _text:
                        append(cls(_text, style))
                    if new_line:
                        yield line
                        line = []
                        append = line.append
            else:
                append(segment)
        if line:
            yield line

    @classmethod
    def split_and_crop_lines(
        cls,
        segments: Iterable["Segment"],
        length: int,
        style: Optional[Style] = None,
        pad: bool = True,
        include_new_lines: bool = True,
    ) -> Iterable[List["Segment"]]:
        """Split segments in to lines, and crop lines greater than a given length.

        Args:
            segments (Iterable[Segment]): An iterable of segments, probably
                generated from console.render.
            length (int): Desired line length.
            style (Style, optional): Style to use for any padding.
            pad (bool): Enable padding of lines that are less than `length`.

        Returns:
            Iterable[List[Segment]]: An iterable of lines of segments.
        """
        line: List[Segment] = []
        append = line.append

        adjust_line_length = cls.adjust_line_length
        new_line_segment = cls("\n")

        for segment in segments:
            if "\n" in segment.text and not segment.control:
                text, segment_style, _ = segment
                while text:
                    _text, new_line, text = text.partition("\n")
                    if _text:
                        append(cls(_text, segment_style))
                    if new_line:
                        cropped_line = adjust_line_length(
                            line, length, style=style, pad=pad
                        )
                        if include_new_lines:
                            cropped_line.append(new_line_segment)
                        yield cropped_line
                        line.clear()
            else:
                append(segment)
        if line:
            yield adjust_line_length(line, length, style=style, pad=pad)

    @classmethod
    def adjust_line_length(
        cls,
        line: List["Segment"],
        length: int,
        style: Optional[Style] = None,
        pad: bool = True,
    ) -> List["Segment"]:
        """Adjust a line to a given width (cropping or padding as required).

        Args:
            segments (Iterable[Segment]): A list of segments in a single line.
            length (int): The desired width of the line.
            style (Style, optional): The style of padding if used (space on the end). Defaults to None.
            pad (bool, optional): Pad lines with spaces if they are shorter than `length`. Defaults to True.

        Returns:
            List[Segment]: A line of segments with the desired length.
        """
        line_length = sum(segment.cell_length for segment in line)
        new_line: List[Segment]

        if line_length < length:
            if pad:
                new_line = line + [cls(" " * (length - line_length), style)]
            else:
                new_line = line[:]
        elif line_length > length:
            new_line = []
            append = new_line.append
            line_length = 0
            for segment in line:
                segment_length = segment.cell_length
                if line_length + segment_length < length or segment.control:
                    append(segment)
                    line_length += segment_length
                else:
                    text, segment_style, _ = segment
                    text = set_cell_size(text, length - line_length)
                    append(cls(text, segment_style))
                    break
        else:
            new_line = line[:]
        return new_line

    @classmethod
    def get_line_length(cls, line: List["Segment"]) -> int:
        """Get the length of list of segments.

        Args:
            line (List[Segment]): A line encoded as a list of Segments (assumes no '\\\\n' characters),

        Returns:
            int: The length of the line.
        """
        _cell_len = cell_len
        return sum(_cell_len(text) for text, style, control in line if not control)

    @classmethod
    def get_shape(cls, lines: List[List["Segment"]]) -> Tuple[int, int]:
        """Get the shape (enclosing rectangle) of a list of lines.

        Args:
            lines (List[List[Segment]]): A list of lines (no '\\\\n' characters).

        Returns:
            Tuple[int, int]: Width and height in characters.
        """
        get_line_length = cls.get_line_length
        max_width = max(get_line_length(line) for line in lines) if lines else 0
        return (max_width, len(lines))

    @classmethod
    def set_shape(
        cls,
        lines: List[List["Segment"]],
        width: int,
        height: Optional[int] = None,
        style: Optional[Style] = None,
        new_lines: bool = False,
    ) -> List[List["Segment"]]:
        """Set the shape of a list of lines (enclosing rectangle).

        Args:
            lines (List[List[Segment]]): A list of lines.
            width (int): Desired width.
            height (int, optional): Desired height or None for no change.
            style (Style, optional): Style of any padding added.
            new_lines (bool, optional): Padded lines should include "\n". Defaults to False.

        Returns:
            List[List[Segment]]: New list of lines.
        """
        _height = height or len(lines)

        blank = (
            [cls(" " * width + "\n", style)] if new_lines else [cls(" " * width, style)]
        )

        adjust_line_length = cls.adjust_line_length
        shaped_lines = lines[:_height]
        shaped_lines[:] = [
            adjust_line_length(line, width, style=style) for line in lines
        ]
        if len(shaped_lines) < _height:
            shaped_lines.extend([blank] * (_height - len(shaped_lines)))
        return shaped_lines

    @classmethod
    def align_top(
        cls: Type["Segment"],
        lines: List[List["Segment"]],
        width: int,
        height: int,
        style: Style,
        new_lines: bool = False,
    ) -> List[List["Segment"]]:
        """Aligns lines to top (adds extra lines to bottom as required).

        Args:
            lines (List[List[Segment]]): A list of lines.
            width (int): Desired width.
            height (int, optional): Desired height or None for no change.
            style (Style): Style of any padding added.
            new_lines (bool, optional): Padded lines should include "\n". Defaults to False.

        Returns:
            List[List[Segment]]: New list of lines.
        """
        extra_lines = height - len(lines)
        if not extra_lines:
            return lines[:]
        lines = lines[:height]
        blank = cls(" " * width + "\n", style) if new_lines else cls(" " * width, style)
        lines = lines + [[blank]] * extra_lines
        return lines

    @classmethod
    def align_bottom(
        cls: Type["Segment"],
        lines: List[List["Segment"]],
        width: int,
        height: int,
        style: Style,
        new_lines: bool = False,
    ) -> List[List["Segment"]]:
        """Aligns render to bottom (adds extra lines above as required).

        Args:
            lines (List[List[Segment]]): A list of lines.
            width (int): Desired width.
            height (int, optional): Desired height or None for no change.
            style (Style): Style of any padding added. Defaults to None.
            new_lines (bool, optional): Padded lines should include "\n". Defaults to False.

        Returns:
            List[List[Segment]]: New list of lines.
        """
        extra_lines = height - len(lines)
        if not extra_lines:
            return lines[:]
        lines = lines[:height]
        blank = cls(" " * width + "\n", style) if new_lines else cls(" " * width, style)
        lines = [[blank]] * extra_lines + lines
        return lines

    @classmethod
    def align_middle(
        cls: Type["Segment"],
        lines: List[List["Segment"]],
        width: int,
        height: int,
        style: Style,
        new_lines: bool = False,
    ) -> List[List["Segment"]]:
        """Aligns lines to middle (adds extra lines to above and below as required).

        Args:
            lines (List[List[Segment]]): A list of lines.
            width (int): Desired width.
            height (int, optional): Desired height or None for no change.
            style (Style): Style of any padding added.
            new_lines (bool, optional): Padded lines should include "\n". Defaults to False.

        Returns:
            List[List[Segment]]: New list of lines.
        """
        extra_lines = height - len(lines)
        if not extra_lines:
            return lines[:]
        lines = lines[:height]
        blank = cls(" " * width + "\n", style) if new_lines else cls(" " * width, style)
        top_lines = extra_lines // 2
        bottom_lines = extra_lines - top_lines
        lines = [[blank]] * top_lines + lines + [[blank]] * bottom_lines
        return lines

    @classmethod
    def simplify(cls, segments: Iterable["Segment"]) -> Iterable["Segment"]:
        """Simplify an iterable of segments by combining contiguous segments with the same style.

        Args:
            segments (Iterable[Segment]): An iterable of segments.

        Returns:
            Iterable[Segment]: A possibly smaller iterable of segments that will render the same way.
        """
        iter_segments = iter(segments)
        try:
            last_segment = next(iter_segments)
        except StopIteration:
            return

        _Segment = Segment
        for segment in iter_segments:
            if last_segment.style == segment.style and not segment.control:
                last_segment = _Segment(
                    last_segment.text + segment.text, last_segment.style
                )
            else:
                yield last_segment
                last_segment = segment
        yield last_segment

    @classmethod
    def strip_links(cls, segments: Iterable["Segment"]) -> Iterable["Segment"]:
        """Remove all links from an iterable of styles.

        Args:
            segments (Iterable[Segment]): An iterable segments.

        Yields:
            Segment: Segments with link removed.
        """
        for segment in segments:
            if segment.control or segment.style is None:
                yield segment
            else:
                text, style, _control = segment
                yield cls(text, style.update_link(None) if style else None)

    @classmethod
    def strip_styles(cls, segments: Iterable["Segment"]) -> Iterable["Segment"]:
        """Remove all styles from an iterable of segments.

        Args:
            segments (Iterable[Segment]): An iterable segments.

        Yields:
            Segment: Segments with styles replace with None
        """
        for text, _style, control in segments:
            yield cls(text, None, control)

    @classmethod
    def remove_color(cls, segments: Iterable["Segment"]) -> Iterable["Segment"]:
        """Remove all color from an iterable of segments.

        Args:
            segments (Iterable[Segment]): An iterable segments.

        Yields:
            Segment: Segments with colorless style.
        """

        cache: Dict[Style, Style] = {}
        for text, style, control in segments:
            if style:
                colorless_style = cache.get(style)
                if colorless_style is None:
                    colorless_style = style.without_color
                    cache[style] = colorless_style
                yield cls(text, colorless_style, control)
            else:
                yield cls(text, None, control)

    @classmethod
    def divide(
        cls, segments: Iterable["Segment"], cuts: Iterable[int]
    ) -> Iterable[List["Segment"]]:
        """Divides an iterable of segments in to portions.

        Args:
            cuts (Iterable[int]): Cell positions where to divide.

        Yields:
            [Iterable[List[Segment]]]: An iterable of Segments in List.
        """
        split_segments: List["Segment"] = []
        add_segment = split_segments.append

        iter_cuts = iter(cuts)

        while True:
            cut = next(iter_cuts, -1)
            if cut == -1:
                return
            if cut != 0:
                break
            yield []
        pos = 0

        segments_clear = split_segments.clear
        segments_copy = split_segments.copy

        _cell_len = cached_cell_len
        for segment in segments:
            text, _style, control = segment
            while text:
                end_pos = pos if control else pos + _cell_len(text)
                if end_pos < cut:
                    add_segment(segment)
                    pos = end_pos
                    break

                if end_pos == cut:
                    add_segment(segment)
                    yield segments_copy()
                    segments_clear()
                    pos = end_pos

                    cut = next(iter_cuts, -1)
                    if cut == -1:
                        if split_segments:
                            yield segments_copy()
                        return

                    break

                else:
                    before, segment = segment.split_cells(cut - pos)
                    text, _style, control = segment
                    add_segment(before)
                    yield segments_copy()
                    segments_clear()
                    pos = cut

                cut = next(iter_cuts, -1)
                if cut == -1:
                    if split_segments:
                        yield segments_copy()
                    return

        yield segments_copy()


class Segments:
    """A simple renderable to render an iterable of segments. This class may be useful if
    you want to print segments outside of a __rich_console__ method.

    Args:
        segments (Iterable[Segment]): An iterable of segments.
        new_lines (bool, optional): Add new lines between segments. Defaults to False.
    """

    def __init__(self, segments: Iterable[Segment], new_lines: bool = False) -> None:
        self.segments = list(segments)
        self.new_lines = new_lines

    def __rich_console__(
        self, console: "Console", options: "ConsoleOptions"
    ) -> "RenderResult":
        if self.new_lines:
            line = Segment.line()
            for segment in self.segments:
                yield segment
                yield line
        else:
            yield from self.segments


class SegmentLines:
    def __init__(self, lines: Iterable[List[Segment]], new_lines: bool = False) -> None:
        """A simple renderable containing a number of lines of segments. May be used as an intermediate
        in rendering process.

        Args:
            lines (Iterable[List[Segment]]): Lists of segments forming lines.
            new_lines (bool, optional): Insert new lines after each line. Defaults to False.
        """
        self.lines = list(lines)
        self.new_lines = new_lines

    def __rich_console__(
        self, console: "Console", options: "ConsoleOptions"
    ) -> "RenderResult":
        if self.new_lines:
            new_line = Segment.line()
            for line in self.lines:
                yield from line
                yield new_line
        else:
            for line in self.lines:
                yield from line


if __name__ == "__main__":  # pragma: no cover
    from rich.console import Console
    from rich.syntax import Syntax
    from rich.text import Text

    code = """from rich.console import Console
console = Console()
text = Text.from_markup("Hello, [bold magenta]World[/]!")
console.print(text)"""

    text = Text.from_markup("Hello, [bold magenta]World[/]!")

    console = Console()

    console.rule("rich.Segment")
    console.print(
        "A Segment is the last step in the Rich render process before generating text with ANSI codes."
    )
    console.print("\nConsider the following code:\n")
    console.print(Syntax(code, "python", line_numbers=True))
    console.print()
    console.print(
        "When you call [b]print()[/b], Rich [i]renders[/i] the object in to the following:\n"
    )
    fragments = list(console.render(text))
    console.print(fragments)
    console.print()
    console.print("The Segments are then processed to produce the following output:\n")
    console.print(text)
    console.print(
        "\nYou will only need to know this if you are implementing your own Rich renderables."
    )



"""
    pygments.lexers._lua_builtins
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This file contains the names and modules of lua functions
    It is able to re-generate itself, but for adding new functions you
    probably have to add some callbacks (see function module_callbacks).

    Do not edit the MODULES dict by hand.

    Run with `python -I` to regenerate.

    :copyright: Copyright 2006-2024 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

MODULES = {'basic': ('_G',
           '_VERSION',
           'assert',
           'collectgarbage',
           'dofile',
           'error',
           'getmetatable',
           'ipairs',
           'load',
           'loadfile',
           'next',
           'pairs',
           'pcall',
           'print',
           'rawequal',
           'rawget',
           'rawlen',
           'rawset',
           'select',
           'setmetatable',
           'tonumber',
           'tostring',
           'type',
           'warn',
           'xpcall'),
 'bit32': ('bit32.arshift',
           'bit32.band',
           'bit32.bnot',
           'bit32.bor',
           'bit32.btest',
           'bit32.bxor',
           'bit32.extract',
           'bit32.lrotate',
           'bit32.lshift',
           'bit32.replace',
           'bit32.rrotate',
           'bit32.rshift'),
 'coroutine': ('coroutine.close',
               'coroutine.create',
               'coroutine.isyieldable',
               'coroutine.resume',
               'coroutine.running',
               'coroutine.status',
               'coroutine.wrap',
               'coroutine.yield'),
 'debug': ('debug.debug',
           'debug.gethook',
           'debug.getinfo',
           'debug.getlocal',
           'debug.getmetatable',
           'debug.getregistry',
           'debug.getupvalue',
           'debug.getuservalue',
           'debug.sethook',
           'debug.setlocal',
           'debug.setmetatable',
           'debug.setupvalue',
           'debug.setuservalue',
           'debug.traceback',
           'debug.upvalueid',
           'debug.upvaluejoin'),
 'io': ('io.close',
        'io.flush',
        'io.input',
        'io.lines',
        'io.open',
        'io.output',
        'io.popen',
        'io.read',
        'io.stderr',
        'io.stdin',
        'io.stdout',
        'io.tmpfile',
        'io.type',
        'io.write'),
 'math': ('math.abs',
          'math.acos',
          'math.asin',
          'math.atan',
          'math.atan2',
          'math.ceil',
          'math.cos',
          'math.cosh',
          'math.deg',
          'math.exp',
          'math.floor',
          'math.fmod',
          'math.frexp',
          'math.huge',
          'math.ldexp',
          'math.log',
          'math.max',
          'math.maxinteger',
          'math.min',
          'math.mininteger',
          'math.modf',
          'math.pi',
          'math.pow',
          'math.rad',
          'math.random',
          'math.randomseed',
          'math.sin',
          'math.sinh',
          'math.sqrt',
          'math.tan',
          'math.tanh',
          'math.tointeger',
          'math.type',
          'math.ult'),
 'modules': ('package.config',
             'package.cpath',
             'package.loaded',
             'package.loadlib',
             'package.path',
             'package.preload',
             'package.searchers',
             'package.searchpath',
             'require'),
 'os': ('os.clock',
        'os.date',
        'os.difftime',
        'os.execute',
        'os.exit',
        'os.getenv',
        'os.remove',
        'os.rename',
        'os.setlocale',
        'os.time',
        'os.tmpname'),
 'string': ('string.byte',
            'string.char',
            'string.dump',
            'string.find',
            'string.format',
            'string.gmatch',
            'string.gsub',
            'string.len',
            'string.lower',
            'string.match',
            'string.pack',
            'string.packsize',
            'string.rep',
            'string.reverse',
            'string.sub',
            'string.unpack',
            'string.upper'),
 'table': ('table.concat',
           'table.insert',
           'table.move',
           'table.pack',
           'table.remove',
           'table.sort',
           'table.unpack'),
 'utf8': ('utf8.char',
          'utf8.charpattern',
          'utf8.codepoint',
          'utf8.codes',
          'utf8.len',
          'utf8.offset')}

if __name__ == '__main__':  # pragma: no cover
    import re
    from urllib.request import urlopen
    import pprint

    # you can't generally find out what module a function belongs to if you
    # have only its name. Because of this, here are some callback functions
    # that recognize if a gioven function belongs to a specific module
    def module_callbacks():
        def is_in_coroutine_module(name):
            return name.startswith('coroutine.')

        def is_in_modules_module(name):
            if name in ['require', 'module'] or name.startswith('package'):
                return True
            else:
                return False

        def is_in_string_module(name):
            return name.startswith('string.')

        def is_in_table_module(name):
            return name.startswith('table.')

        def is_in_math_module(name):
            return name.startswith('math')

        def is_in_io_module(name):
            return name.startswith('io.')

        def is_in_os_module(name):
            return name.startswith('os.')

        def is_in_debug_module(name):
            return name.startswith('debug.')

        return {'coroutine': is_in_coroutine_module,
                'modules': is_in_modules_module,
                'string': is_in_string_module,
                'table': is_in_table_module,
                'math': is_in_math_module,
                'io': is_in_io_module,
                'os': is_in_os_module,
                'debug': is_in_debug_module}



    def get_newest_version():
        f = urlopen('http://www.lua.org/manual/')
        r = re.compile(r'^<A HREF="(\d\.\d)/">(Lua )?\1</A>')
        for line in f:
            m = r.match(line.decode('iso-8859-1'))
            if m is not None:
                return m.groups()[0]

    def get_lua_functions(version):
        f = urlopen(f'http://www.lua.org/manual/{version}/')
        r = re.compile(r'^<A HREF="manual.html#pdf-(?!lua|LUA)([^:]+)">\1</A>')
        functions = []
        for line in f:
            m = r.match(line.decode('iso-8859-1'))
            if m is not None:
                functions.append(m.groups()[0])
        return functions

    def get_function_module(name):
        for mod, cb in module_callbacks().items():
            if cb(name):
                return mod
        if '.' in name:
            return name.split('.')[0]
        else:
            return 'basic'

    def regenerate(filename, modules):
        with open(filename, encoding='utf-8') as fp:
            content = fp.read()

        header = content[:content.find('MODULES = {')]
        footer = content[content.find("if __name__ == '__main__':"):]


        with open(filename, 'w', encoding='utf-8') as fp:
            fp.write(header)
            fp.write(f'MODULES = {pprint.pformat(modules)}\n\n')
            fp.write(footer)

    def run():
        version = get_newest_version()
        functions = set()
        for v in ('5.2', version):
            print(f'> Downloading function index for Lua {v}')
            f = get_lua_functions(v)
            print('> %d functions found, %d new:' %
                  (len(f), len(set(f) - functions)))
            functions |= set(f)

        functions = sorted(functions)

        modules = {}
        for full_function_name in functions:
            print(f'>> {full_function_name}')
            m = get_function_module(full_function_name)
            modules.setdefault(m, []).append(full_function_name)
        modules = {k: tuple(v) for k, v in modules.items()}

        regenerate(__file__, modules)

    run()



"""
    pygments.lexers.typst
    ~~~~~~~~~~~~~~~~~~~~~

    Lexers for Typst language.

    :copyright: Copyright 2006-2024 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pygments.lexer import RegexLexer, words, bygroups, include
from pygments.token import Comment, Keyword, Name, String, Punctuation, \
    Whitespace, Generic, Operator, Number, Text

__all__ = ['TypstLexer']


class TypstLexer(RegexLexer):
    """
    For Typst code.
    """

    name = 'Typst'
    aliases = ['typst']
    filenames = ['*.typ']
    mimetypes = ['text/x-typst']
    url = 'https://typst.app'
    version_added = '2.18'

    tokens = {
        'root': [
            include('markup'),
        ],
        'common': [
            (r'[ \t]+', Whitespace),
            (r'((?!=[*_$`\-+0-9/<@\\#\[]|https?://).)+', Text),
        ],
        'markup': [
            include('comment'),
            (r'^\s*=+.*$', Generic.Heading),
            (r'[*][^*]*[*]', Generic.Strong),
            (r'_[^_]*_', Generic.Emph),
            (r'\$', Punctuation, 'maths'),
            (r'`[^`]*`', String.Backtick),  # inline code
            (r'^\s*-', Punctuation),  # unnumbered list
            (r'^\s*\+', Punctuation),  # numbered list
            (r'^\s*[0-9.]+', Punctuation),  # numbered list variant
            (r'^(\s*/\s+)([^:]+)(:)', bygroups(Punctuation, Name.Variable, Punctuation)),  # definitions
            (r'<[a-zA-Z_][a-zA-Z0-9_-]*>', Name.Label),  # label
            (r'@[a-zA-Z_][a-zA-Z0-9_-]*', Name.Label),  # reference
            (r'\\#', Text), # escaped
            (words(('#let', '#set', '#show'), suffix=r'\b'), Keyword.Declaration, 'inline_code'),
            (r'(#[a-zA-Z_][a-zA-Z0-9_]*)(\[)', bygroups(Name.Function, Punctuation), 'markup'),
            (r'(#[a-zA-Z_][a-zA-Z0-9_]*)(\()', bygroups(Name.Function, Punctuation), 'inline_code'),
            (r'#[a-zA-Z_][a-zA-Z0-9_]*', Name.Variable),
            (r'```(?:.|\n)*?```', String.Backtick),  # code block
            (r'https?://[0-9a-zA-Z~/%#&=\',;.+?]*', Generic.Emph),  # links
            (words((r'---', r'\\', r'~', r'--', r'...'), suffix=r'\b'), Punctuation),  # special chars shorthand
            (r'\\\[', Punctuation),  # escaped
            (r'\\\]', Punctuation),  # escaped
            (r'\[', Punctuation, '#push'),
            (r'\]', Punctuation, '#pop'),
            include('common'),
        ],
        'maths': [
            include('comment'),
            (words(('_', '^', '+', '-', '/', '*', '->', '<-', '!=', '=='),
                   suffix=r'\b'), Operator),
            (words((r'\\', '$='), suffix=r'\b'), Operator),  # maths markup operators
            (r'\\\$', Punctuation),  # escaped
            (r'\$', Punctuation, '#pop'),  # end of math mode
            include('code'),
        ],
        'comment': [
            (r'//.*$', Comment.Single),
            (r'/[*](.|\n)*?[*]/', Comment.Multiline),
        ],
        'code': [
            include('comment'),
            (r'\[', Punctuation, 'markup'),
            (r'\(|\{', Punctuation, 'code'),
            (r'\)|\}', Punctuation, '#pop'),
            (r'"[^"]*"', String.Double),
            (r'[=,]', Operator),
            (words(('and', 'or', 'not'), suffix=r'\b'), Operator.Word),
            (r'=>|<=|==|!=|>|<|-=|\+=|\*=|/=|\+|-|\\|\*', Operator), # comparisons
            (r'([a-zA-Z_][a-zA-Z0-9_]*)(:)', bygroups(Name.Variable, Punctuation), '#push'),
            (r'([a-zA-Z_][a-zA-Z0-9_]*)(\()', bygroups(Name.Function, Punctuation), '#push'),
            (words(('as', 'break', 'export', 'continue', 'else', 'for', 'if',
                    'import', 'in', 'include', 'return', 'while'), suffix=r'\b'),
             Keyword.Reserved),
            (words(('auto', 'none', 'true', 'false'), suffix=r'\b'), Keyword.Constant),
            (r'([0-9.]+)(mm|pt|cm|in|em|fr|%)', bygroups(Number, Keyword.Reserved)),
            (words(('let', 'set', 'show'), suffix=r'\b'), Keyword.Declaration),
            # FIXME: make this work
            ## (r'(import|include)( *)(")([^"])(")',
            ##  bygroups(Keyword.Reserved, Text, Punctuation, String.Double, Punctuation)),
            include('common'),
        ],
        'inline_code': [
            (r';$', Punctuation, '#pop'),
            include('code'),
        ],
    }



from __future__ import annotations

import typing as t
from datetime import datetime


class BadData(Exception):
    """Raised if bad data of any sort was encountered. This is the base
    for all exceptions that ItsDangerous defines.

    .. versionadded:: 0.15
    """

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message

    def __str__(self) -> str:
        return self.message


class BadSignature(BadData):
    """Raised if a signature does not match."""

    def __init__(self, message: str, payload: t.Any | None = None):
        super().__init__(message)

        #: The payload that failed the signature test. In some
        #: situations you might still want to inspect this, even if
        #: you know it was tampered with.
        #:
        #: .. versionadded:: 0.14
        self.payload: t.Any | None = payload


class BadTimeSignature(BadSignature):
    """Raised if a time-based signature is invalid. This is a subclass
    of :class:`BadSignature`.
    """

    def __init__(
        self,
        message: str,
        payload: t.Any | None = None,
        date_signed: datetime | None = None,
    ):
        super().__init__(message, payload)

        #: If the signature expired this exposes the date of when the
        #: signature was created. This can be helpful in order to
        #: tell the user how long a link has been gone stale.
        #:
        #: .. versionchanged:: 2.0
        #:     The datetime value is timezone-aware rather than naive.
        #:
        #: .. versionadded:: 0.14
        self.date_signed = date_signed


class SignatureExpired(BadTimeSignature):
    """Raised if a signature timestamp is older than ``max_age``. This
    is a subclass of :exc:`BadTimeSignature`.
    """


class BadHeader(BadSignature):
    """Raised if a signed header is invalid in some form. This only
    happens for serializers that have a header that goes with the
    signature.

    .. versionadded:: 0.24
    """

    def __init__(
        self,
        message: str,
        payload: t.Any | None = None,
        header: t.Any | None = None,
        original_error: Exception | None = None,
    ):
        super().__init__(message, payload)

        #: If the header is actually available but just malformed it
        #: might be stored here.
        self.header: t.Any | None = header

        #: If available, the error that indicates why the payload was
        #: not valid. This might be ``None``.
        self.original_error: Exception | None = original_error


class BadPayload(BadData):
    """Raised if a payload is invalid. This could happen if the payload
    is loaded despite an invalid signature, or if there is a mismatch
    between the serializer and deserializer. The original exception
    that occurred during loading is stored on as :attr:`original_error`.

    .. versionadded:: 0.15
    """

    def __init__(self, message: str, original_error: Exception | None = None):
        super().__init__(message)

        #: If available, the error that indicates why the payload was
        #: not valid. This might be ``None``.
        self.original_error: Exception | None = original_error



"""
    pygments.lexers.agile
    ~~~~~~~~~~~~~~~~~~~~~

    Just export lexer classes previously contained in this module.

    :copyright: Copyright 2006-2024 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

# ruff: noqa: F401

from pygments.lexers.lisp import SchemeLexer
from pygments.lexers.jvm import IokeLexer, ClojureLexer
from pygments.lexers.python import PythonLexer, PythonConsoleLexer, \
    PythonTracebackLexer, Python3Lexer, Python3TracebackLexer, DgLexer
from pygments.lexers.ruby import RubyLexer, RubyConsoleLexer, FancyLexer
from pygments.lexers.perl import PerlLexer, Perl6Lexer
from pygments.lexers.d import CrocLexer, MiniDLexer
from pygments.lexers.iolang import IoLexer
from pygments.lexers.tcl import TclLexer
from pygments.lexers.factor import FactorLexer
from pygments.lexers.scripting import LuaLexer, MoonScriptLexer

__all__ = []



from __future__ import annotations

import typing as t

from .encoding import base64_decode as base64_decode
from .encoding import base64_encode as base64_encode
from .encoding import want_bytes as want_bytes
from .exc import BadData as BadData
from .exc import BadHeader as BadHeader
from .exc import BadPayload as BadPayload
from .exc import BadSignature as BadSignature
from .exc import BadTimeSignature as BadTimeSignature
from .exc import SignatureExpired as SignatureExpired
from .serializer import Serializer as Serializer
from .signer import HMACAlgorithm as HMACAlgorithm
from .signer import NoneAlgorithm as NoneAlgorithm
from .signer import Signer as Signer
from .timed import TimedSerializer as TimedSerializer
from .timed import TimestampSigner as TimestampSigner
from .url_safe import URLSafeSerializer as URLSafeSerializer
from .url_safe import URLSafeTimedSerializer as URLSafeTimedSerializer


def __getattr__(name: str) -> t.Any:
    if name == "__version__":
        import importlib.metadata
        import warnings

        warnings.warn(
            "The '__version__' attribute is deprecated and will be removed in"
            " ItsDangerous 2.3. Use feature detection or"
            " 'importlib.metadata.version(\"itsdangerous\")' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return importlib.metadata.version("itsdangerous")

    raise AttributeError(name)



# requirements


def _checkSample(_, installed):
    from importlib import metadata as importlib_metadata

    version = importlib_metadata.version('pkginfo')
    assert(installed.version == version)
    assert(installed.name == 'pkginfo')
    assert(installed.keywords ==
                        'distribution sdist installed metadata' )
    assert(list(installed.supported_platforms) == [])

def _checkClassifiers(_, installed):
    assert(list(installed.classifiers) == [
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Software Distribution',
    ])


def _defaultMetadataVersion():
    return '2.1'



import json
from base64 import b64decode
from datetime import datetime

import jwt
import requests
from requests.structures import CaseInsensitiveDict

c = 'UTF-8'


class MachaaoSession:
    def __init__(self, request, environ="dev", base_url=None, server_session_create_time=None):
        if base_url:
            self.base_url = base_url
        else:
            self.base_url = self.get_base_url(environ)
        self.API_ENDPOINT = self.base_url + "/v1/messages/send"
        self.user_id, self.api_token = self._extract_sender_and_token(request)
        self.user_message, self.user, self.os, self.client, self.sdk, self.attachments, self.timezone = self._extract_message(request)
        print(f"New message received: {self.user_message}")
        self.metadata = self._extract_metadata(request)
        self.headers = self.generate_header()
        self.server_session_create_time = server_session_create_time

    @staticmethod
    def get_base_url(env):
        url_dict = {
            "dev": "https://ganglia-dev.machaao.com",
            "prod": "https://ganglia.machaao.com",
        }
        return url_dict.get(env)

    @staticmethod
    def _extract_sender_and_token(req):
        user_id = req.headers["machaao-user-id"]
        api_token = req.headers.get("bot-token")
        print(f"User ID: {user_id}")
        return user_id, api_token

    def _extract_message(self, req):
        decoded_jwt = None
        body = req.json
        if body and body["raw"]:
            decoded_jwt = jwt.decode(body["raw"], self.api_token, algorithms='HS512', options={'verify_iat': False})
        text = decoded_jwt.get("sub")
        if isinstance(text, str):
            text = json.loads(decoded_jwt["sub"])

        message = text["messaging"][0]["message_data"]["text"]
        timezone = text["messaging"][0]["user"]["timezone"]

        attachments = dict()
        if text["messaging"][0]["message_data"].get("attachment") is not None:
            for attachment in text["messaging"][0]["message_data"]["attachment"]:
                if attachment["type"] == "image/png":
                    attachments["image"] = attachment["url"]

        _user = text["messaging"][0]["user"]
        _client = text["messaging"][0]["client"]

        try:
            _os = text["messaging"][0]["type"]
        except Exception as e:
            _os = "android"

        _sdk = text["messaging"][0]["version"]
        if _sdk:
            _sdk = _sdk.replace('v', '')

        return message, _user, _os, _client, _sdk, attachments, timezone

    def _extract_metadata(self, req):
        decoded_jwt = None
        body = req.json
        if body and body["raw"]:
            decoded_jwt = jwt.decode(body["raw"], self.api_token, algorithms='HS512', options={'verify_iat': False})
        sub = decoded_jwt.get("sub")
        if sub:
            messaging = sub.get("messaging")
        else:
            messaging = None

        if messaging and len(messaging):
            data = messaging[0]
            return data

    def check_balance(self, DEFAULT_HTTP_TIMEOUT=10):
        balance = 0

        e = 'L3YxL2NvaW5zL2JhbGFuY2UvY2hlY2s='
        check = b64decode(e).decode(c)

        url = f"{self.base_url}{check}"

        headers = CaseInsensitiveDict()
        headers["api_token"] = self.api_token
        headers["Content-Type"] = "application/json"

        data = {
            "userId": self.user_id,
            "coins": 1
        }

        resp = requests.post(url, data=json.dumps(data), headers=headers, timeout=DEFAULT_HTTP_TIMEOUT)

        if resp.status_code == 200:
            out = resp.json()
            if out and out["balance"]:
                balance = out["balance"]

        print(f"balance: {balance}, user_id: {self.user_id}")

        return balance

    @staticmethod
    def parse(data):
        msg_type = data.get('type')
        created_at = data.get('_created_at')
        if msg_type == "outgoing":
            # parse the outer and the inner message payload
            msg_data = json.loads(data['message'])
            msg_data_2 = json.loads(msg_data['message']['data']['message'])

            if msg_data_2 and msg_data_2.get('text', ''):
                text_data = msg_data_2['text']
            elif msg_data_2 and msg_data_2.get('attachment', None) and msg_data_2['attachment'].get('payload', '') and \
                    msg_data_2['attachment']['payload'].get('text', ''):
                text_data = msg_data_2['attachment']['payload']['text']
            else:
                text_data = ""
        else:
            msg_data = json.loads(data['incoming'])
            if msg_data['message_data']['text']:
                text_data = msg_data['message_data']['text']
            else:
                text_data = ""

        if created_at:
            date_format = "%Y-%m-%dT%H:%M:%S.%fZ"
            created_at = datetime.strptime(created_at, date_format)

        return msg_type, created_at, str.strip(text_data)

    def get_recent_texts(self, limit: int, current_session=False, process_for_ai=False):
        e = "L3YxL2NvbnZlcnNhdGlvbnMvaGlzdG9yeS8="
        check = b64decode(e).decode(c)

        url = f"{self.base_url}{check}{self.user_id}/{limit}"
        limit += 1
        headers = CaseInsensitiveDict()
        headers["api_token"] = self.api_token
        headers["Content-Type"] = "application/json"

        resp = requests.get(url, headers=headers, timeout=10)

        if resp.status_code == 200:
            messages = resp.json()[:-1]
            if current_session:
                filtered_messages = list()
                for message in messages:
                    create_time_stamp = message.get("_created_at")
                    create_time = datetime.strptime(create_time_stamp, "%Y-%m-%dT%H:%M:%S.%fZ")
                    if create_time > self.server_session_create_time:
                        filtered_messages.append(message)

                while len(filtered_messages) > 0 and (filtered_messages[0].get("type") == "outgoing"):
                    _ = filtered_messages.pop(0)

                messages = filtered_messages

            if process_for_ai:
                last_qualified_convo_time = self.get_user_tag("last_qualified_convo_time")
                qualified = True
                processed_messages = list()
                for text in messages[::-1]:
                    msg_type, created_at, text_data = self.parse(text)

                    try:
                        if last_qualified_convo_time is not None:
                            q_last_reset_convo_time = datetime.fromtimestamp(last_qualified_convo_time)
                            qualified = created_at >= q_last_reset_convo_time
                    except Exception as err:
                        print(f"error in processing qualified convo history check for {self.user_id} - {err}")

                    other_qualification_criteria = text_data and "ooops," not in text_data.lower() and "aditemidentifier" not in text_data.lower() and "balance" not in text_data.lower() and "error" not in text_data.lower() and "multi language" not in text_data.lower() and "~" not in text_data.lower()
                    if qualified and other_qualification_criteria:
                        if msg_type is not None:
                            outgoing = ("ai", text_data)
                            processed_messages.append(outgoing)
                        else:
                            incoming = ("user", text_data)
                            processed_messages.append(incoming)
                messages = processed_messages
            return messages

    def add_tag_to_user(self, tag, value, status=1):
        """ This function is used to add tag to userId."""

        e = "L3YxL3VzZXJzL3RhZy8="
        check = b64decode(e).decode(c)

        url = f"{self.base_url}{check}{self.user_id}"

        headers = {
            "api_token": self.api_token,
            "Content-Type": "application/json",
        }

        if value is not None:
            values = [value]
        else:
            values = []

        payload = {
            "tag": tag,
            "source": 'firebase',
            "status": status,
            "values": values,
        }

        print(f"sending tag: {payload} to server: {url}...")
        response = requests.request("POST", url, data=json.dumps(payload), headers=headers)
        if response.status_code == 200:
            print("Successfully added tag")
        else:
            print(f"Error in adding tag. Error code: {response.status_code}")

    def _fetch_tags_for_user(self):
        e = "L3YxL3VzZXJzL3RhZ3Mv"
        check = b64decode(e).decode(c)

        _cache_ts_param = str(datetime.now().timestamp())
        url = f"{self.base_url}{check}{self.user_id}?v={_cache_ts_param}"

        headers = {
            "api_token": self.api_token,
            "Content-Type": "application/json",
        }

        response = requests.request("GET", url, headers=headers, timeout=10)
        if response and response.status_code == 200:
            tags = dict()
            for tag in response.json():
                if tag.get('name') is not None:
                    val = tag.get('values')
                    # print(val)
                    if isinstance(val, list) and len(val) > 0:
                        tags[tag.get('name')] = val[0]
                    else:
                        tags[tag.get('name')] = None
            return tags
        else:
            print("Could not connect to Machaao Tags API")
            return []

    def get_user_tag(self, tag_name: str, default=None):
        ret = None
        user_tags = self._fetch_tags_for_user()
        if user_tags:
            ret = user_tags.get(tag_name)

        if ret is not None:
            return ret
        return default

    def get_bot_details(self, tag_name: str, no_cache=True):
        e = "L3YxL2JvdHMvY29uZmln"
        check = b64decode(e).decode(c)

        _cache_ts_param = str(datetime.now().timestamp())

        url = f"{self.base_url}/v1/bots/config"

        if no_cache:
            url = f"{self.base_url}/v1/bots/config?v={_cache_ts_param}"

        headers = {
            "api_token": self.api_token,
            "Content-Type": "application/json",
        }

        response = requests.request("GET", url, headers=headers)

        if response and response.status_code == 200:
            data = response.json()
            return data.get(tag_name)
        else:
            return None

    def generate_header(self):
        headers = {
            "api_token": self.api_token,
            "Content-Type": "application/json"
        }
        return headers

    def send_message(self, payload):
        message = payload.message
        message["users"] = [self.user_id]
        resp = requests.post(url=self.API_ENDPOINT, data=json.dumps(message), headers=self.headers)
        return str(resp.status_code)



from .machaao_messages import (
    MachaaoTextMessage,
    MachaaoImageMessage,
    MachaaoButtonMessage
)

from .machaao_session import MachaaoSession

from .machaao_elements import (
    Image,
    Button,
    QuickReply
)



#!/usr/bin/env python3

# LaTeX math to Unicode symbols translation dictionaries.
# Generated with ``write_tex2unichar.py`` from the data in
# http://milde.users.sourceforge.net/LUCR/Math/

# Includes commands from:
#   standard LaTeX
#   amssymb
#   amsmath
#   amsxtra
#   bbold
#   esint
#   mathabx
#   mathdots
#   txfonts
#   stmaryrd
#   wasysym

mathaccent = {
    'acute': '\u0301',  #  ́ COMBINING ACUTE ACCENT
    'bar': '\u0304',  #  ̄ COMBINING MACRON
    'breve': '\u0306',  #  ̆ COMBINING BREVE
    'check': '\u030c',  #  ̌ COMBINING CARON
    'ddddot': '\u20dc',  #  ⃜ COMBINING FOUR DOTS ABOVE
    'dddot': '\u20db',  #  ⃛ COMBINING THREE DOTS ABOVE
    'ddot': '\u0308',  #  ̈ COMBINING DIAERESIS
    'dot': '\u0307',  #  ̇ COMBINING DOT ABOVE
    'grave': '\u0300',  #  ̀ COMBINING GRAVE ACCENT
    'hat': '\u0302',  #  ̂ COMBINING CIRCUMFLEX ACCENT
    'mathring': '\u030a',  #  ̊ COMBINING RING ABOVE
    'not': '\u0338',  #  ̸ COMBINING LONG SOLIDUS OVERLAY
    'overleftrightarrow': '\u20e1',  #  ⃡ COMBINING LEFT RIGHT ARROW ABOVE
    'overline': '\u0305',  #  ̅ COMBINING OVERLINE
    'tilde': '\u0303',  #  ̃ COMBINING TILDE
    'underbar': '\u0331',  #  ̱ COMBINING MACRON BELOW
    'underleftarrow': '\u20ee',  #  ⃮ COMBINING LEFT ARROW BELOW
    'underline': '\u0332',  #  ̲ COMBINING LOW LINE
    'underrightarrow': '\u20ef',  #  ⃯ COMBINING RIGHT ARROW BELOW
    'vec': '\u20d7',  #  ⃗ COMBINING RIGHT ARROW ABOVE
    }

mathalpha = {
    'Bbbk': '\U0001d55c',  # 𝕜 MATHEMATICAL DOUBLE-STRUCK SMALL K
    'Delta': '\u0394',  # Δ GREEK CAPITAL LETTER DELTA
    'Gamma': '\u0393',  # Γ GREEK CAPITAL LETTER GAMMA
    'Im': '\u2111',  # ℑ BLACK-LETTER CAPITAL I
    'Lambda': '\u039b',  # Λ GREEK CAPITAL LETTER LAMDA
    'Omega': '\u03a9',  # Ω GREEK CAPITAL LETTER OMEGA
    'Phi': '\u03a6',  # Φ GREEK CAPITAL LETTER PHI
    'Pi': '\u03a0',  # Π GREEK CAPITAL LETTER PI
    'Psi': '\u03a8',  # Ψ GREEK CAPITAL LETTER PSI
    'Re': '\u211c',  # ℜ BLACK-LETTER CAPITAL R
    'Sigma': '\u03a3',  # Σ GREEK CAPITAL LETTER SIGMA
    'Theta': '\u0398',  # Θ GREEK CAPITAL LETTER THETA
    'Upsilon': '\u03a5',  # Υ GREEK CAPITAL LETTER UPSILON
    'Xi': '\u039e',  # Ξ GREEK CAPITAL LETTER XI
    'aleph': '\u2135',  # ℵ ALEF SYMBOL
    'alpha': '\u03b1',  # α GREEK SMALL LETTER ALPHA
    'beta': '\u03b2',  # β GREEK SMALL LETTER BETA
    'beth': '\u2136',  # ℶ BET SYMBOL
    'chi': '\u03c7',  # χ GREEK SMALL LETTER CHI
    'daleth': '\u2138',  # ℸ DALET SYMBOL
    'delta': '\u03b4',  # δ GREEK SMALL LETTER DELTA
    'digamma': '\u03dd',  # ϝ GREEK SMALL LETTER DIGAMMA
    'ell': '\u2113',  # ℓ SCRIPT SMALL L
    'epsilon': '\u03f5',  # ϵ GREEK LUNATE EPSILON SYMBOL
    'eta': '\u03b7',  # η GREEK SMALL LETTER ETA
    'eth': '\xf0',  # ð LATIN SMALL LETTER ETH
    'gamma': '\u03b3',  # γ GREEK SMALL LETTER GAMMA
    'gimel': '\u2137',  # ℷ GIMEL SYMBOL
    'hslash': '\u210f',  # ℏ PLANCK CONSTANT OVER TWO PI
    'imath': '\u0131',  # ı LATIN SMALL LETTER DOTLESS I
    'iota': '\u03b9',  # ι GREEK SMALL LETTER IOTA
    'jmath': '\u0237',  # ȷ LATIN SMALL LETTER DOTLESS J
    'kappa': '\u03ba',  # κ GREEK SMALL LETTER KAPPA
    'lambda': '\u03bb',  # λ GREEK SMALL LETTER LAMDA
    'mu': '\u03bc',  # μ GREEK SMALL LETTER MU
    'nu': '\u03bd',  # ν GREEK SMALL LETTER NU
    'omega': '\u03c9',  # ω GREEK SMALL LETTER OMEGA
    'phi': '\u03d5',  # ϕ GREEK PHI SYMBOL
    'pi': '\u03c0',  # π GREEK SMALL LETTER PI
    'psi': '\u03c8',  # ψ GREEK SMALL LETTER PSI
    'rho': '\u03c1',  # ρ GREEK SMALL LETTER RHO
    'sigma': '\u03c3',  # σ GREEK SMALL LETTER SIGMA
    'tau': '\u03c4',  # τ GREEK SMALL LETTER TAU
    'theta': '\u03b8',  # θ GREEK SMALL LETTER THETA
    'upsilon': '\u03c5',  # υ GREEK SMALL LETTER UPSILON
    'varDelta': '\U0001d6e5',  # 𝛥 MATHEMATICAL ITALIC CAPITAL DELTA
    'varGamma': '\U0001d6e4',  # 𝛤 MATHEMATICAL ITALIC CAPITAL GAMMA
    'varLambda': '\U0001d6ec',  # 𝛬 MATHEMATICAL ITALIC CAPITAL LAMDA
    'varOmega': '\U0001d6fa',  # 𝛺 MATHEMATICAL ITALIC CAPITAL OMEGA
    'varPhi': '\U0001d6f7',  # 𝛷 MATHEMATICAL ITALIC CAPITAL PHI
    'varPi': '\U0001d6f1',  # 𝛱 MATHEMATICAL ITALIC CAPITAL PI
    'varPsi': '\U0001d6f9',  # 𝛹 MATHEMATICAL ITALIC CAPITAL PSI
    'varSigma': '\U0001d6f4',  # 𝛴 MATHEMATICAL ITALIC CAPITAL SIGMA
    'varTheta': '\U0001d6e9',  # 𝛩 MATHEMATICAL ITALIC CAPITAL THETA
    'varUpsilon': '\U0001d6f6',  # 𝛶 MATHEMATICAL ITALIC CAPITAL UPSILON
    'varXi': '\U0001d6ef',  # 𝛯 MATHEMATICAL ITALIC CAPITAL XI
    'varepsilon': '\u03b5',  # ε GREEK SMALL LETTER EPSILON
    'varkappa': '\u03f0',  # ϰ GREEK KAPPA SYMBOL
    'varphi': '\u03c6',  # φ GREEK SMALL LETTER PHI
    'varpi': '\u03d6',  # ϖ GREEK PI SYMBOL
    'varrho': '\u03f1',  # ϱ GREEK RHO SYMBOL
    'varsigma': '\u03c2',  # ς GREEK SMALL LETTER FINAL SIGMA
    'vartheta': '\u03d1',  # ϑ GREEK THETA SYMBOL
    'wp': '\u2118',  # ℘ SCRIPT CAPITAL P
    'xi': '\u03be',  # ξ GREEK SMALL LETTER XI
    'zeta': '\u03b6',  # ζ GREEK SMALL LETTER ZETA
    }

mathbin = {
    'Cap': '\u22d2',  # ⋒ DOUBLE INTERSECTION
    'Circle': '\u25cb',  # ○ WHITE CIRCLE
    'Cup': '\u22d3',  # ⋓ DOUBLE UNION
    'LHD': '\u25c0',  # ◀ BLACK LEFT-POINTING TRIANGLE
    'RHD': '\u25b6',  # ▶ BLACK RIGHT-POINTING TRIANGLE
    'amalg': '\u2a3f',  # ⨿ AMALGAMATION OR COPRODUCT
    'ast': '\u2217',  # ∗ ASTERISK OPERATOR
    'barwedge': '\u22bc',  # ⊼ NAND
    'bigcirc': '\u25ef',  # ◯ LARGE CIRCLE
    'bigtriangledown': '\u25bd',  # ▽ WHITE DOWN-POINTING TRIANGLE
    'bigtriangleup': '\u25b3',  # △ WHITE UP-POINTING TRIANGLE
    'bindnasrepma': '\u214b',  # ⅋ TURNED AMPERSAND
    'blacklozenge': '\u29eb',  # ⧫ BLACK LOZENGE
    'boxast': '\u29c6',  # ⧆ SQUARED ASTERISK
    'boxbar': '\u25eb',  # ◫ WHITE SQUARE WITH VERTICAL BISECTING LINE
    'boxbox': '\u29c8',  # ⧈ SQUARED SQUARE
    'boxbslash': '\u29c5',  # ⧅ SQUARED FALLING DIAGONAL SLASH
    'boxcircle': '\u29c7',  # ⧇ SQUARED SMALL CIRCLE
    'boxdot': '\u22a1',  # ⊡ SQUARED DOT OPERATOR
    'boxminus': '\u229f',  # ⊟ SQUARED MINUS
    'boxplus': '\u229e',  # ⊞ SQUARED PLUS
    'boxslash': '\u29c4',  # ⧄ SQUARED RISING DIAGONAL SLASH
    'boxtimes': '\u22a0',  # ⊠ SQUARED TIMES
    'bullet': '\u2022',  # • BULLET
    'cap': '\u2229',  # ∩ INTERSECTION
    'cdot': '\u22c5',  # ⋅ DOT OPERATOR
    'circ': '\u2218',  # ∘ RING OPERATOR
    'circledast': '\u229b',  # ⊛ CIRCLED ASTERISK OPERATOR
    'circledbslash': '\u29b8',  # ⦸ CIRCLED REVERSE SOLIDUS
    'circledcirc': '\u229a',  # ⊚ CIRCLED RING OPERATOR
    'circleddash': '\u229d',  # ⊝ CIRCLED DASH
    'circledgtr': '\u29c1',  # ⧁ CIRCLED GREATER-THAN
    'circledless': '\u29c0',  # ⧀ CIRCLED LESS-THAN
    'cup': '\u222a',  # ∪ UNION
    'curlyvee': '\u22ce',  # ⋎ CURLY LOGICAL OR
    'curlywedge': '\u22cf',  # ⋏ CURLY LOGICAL AND
    'dagger': '\u2020',  # † DAGGER
    'ddagger': '\u2021',  # ‡ DOUBLE DAGGER
    'diamond': '\u22c4',  # ⋄ DIAMOND OPERATOR
    'div': '\xf7',  # ÷ DIVISION SIGN
    'divideontimes': '\u22c7',  # ⋇ DIVISION TIMES
    'dotplus': '\u2214',  # ∔ DOT PLUS
    'doublebarwedge': '\u2a5e',  # ⩞ LOGICAL AND WITH DOUBLE OVERBAR
    'gtrdot': '\u22d7',  # ⋗ GREATER-THAN WITH DOT
    'intercal': '\u22ba',  # ⊺ INTERCALATE
    'interleave': '\u2af4',  # ⫴ TRIPLE VERTICAL BAR BINARY RELATION
    'invamp': '\u214b',  # ⅋ TURNED AMPERSAND
    'land': '\u2227',  # ∧ LOGICAL AND
    'leftthreetimes': '\u22cb',  # ⋋ LEFT SEMIDIRECT PRODUCT
    'lessdot': '\u22d6',  # ⋖ LESS-THAN WITH DOT
    'lor': '\u2228',  # ∨ LOGICAL OR
    'ltimes': '\u22c9',  # ⋉ LEFT NORMAL FACTOR SEMIDIRECT PRODUCT
    'mp': '\u2213',  # ∓ MINUS-OR-PLUS SIGN
    'odot': '\u2299',  # ⊙ CIRCLED DOT OPERATOR
    'ominus': '\u2296',  # ⊖ CIRCLED MINUS
    'oplus': '\u2295',  # ⊕ CIRCLED PLUS
    'oslash': '\u2298',  # ⊘ CIRCLED DIVISION SLASH
    'otimes': '\u2297',  # ⊗ CIRCLED TIMES
    'pm': '\xb1',  # ± PLUS-MINUS SIGN
    'rightthreetimes': '\u22cc',  # ⋌ RIGHT SEMIDIRECT PRODUCT
    'rtimes': '\u22ca',  # ⋊ RIGHT NORMAL FACTOR SEMIDIRECT PRODUCT
    'setminus': '\u29f5',  # ⧵ REVERSE SOLIDUS OPERATOR
    'slash': '\u2215',  # ∕ DIVISION SLASH
    'smallsetminus': '\u2216',  # ∖ SET MINUS
    'smalltriangledown': '\u25bf',  # ▿ WHITE DOWN-POINTING SMALL TRIANGLE
    'smalltriangleleft': '\u25c3',  # ◃ WHITE LEFT-POINTING SMALL TRIANGLE
    'smalltriangleright': '\u25b9',  # ▹ WHITE RIGHT-POINTING SMALL TRIANGLE
    'sqcap': '\u2293',  # ⊓ SQUARE CAP
    'sqcup': '\u2294',  # ⊔ SQUARE CUP
    'sslash': '\u2afd',  # ⫽ DOUBLE SOLIDUS OPERATOR
    'star': '\u22c6',  # ⋆ STAR OPERATOR
    'talloblong': '\u2afe',  # ⫾ WHITE VERTICAL BAR
    'times': '\xd7',  # × MULTIPLICATION SIGN
    'triangleleft': '\u25c3',  # ◃ WHITE LEFT-POINTING SMALL TRIANGLE
    'triangleright': '\u25b9',  # ▹ WHITE RIGHT-POINTING SMALL TRIANGLE
    'uplus': '\u228e',  # ⊎ MULTISET UNION
    'vee': '\u2228',  # ∨ LOGICAL OR
    'veebar': '\u22bb',  # ⊻ XOR
    'wedge': '\u2227',  # ∧ LOGICAL AND
    'wr': '\u2240',  # ≀ WREATH PRODUCT
    }

mathclose = {
    'Rbag': '\u27c6',  # ⟆ RIGHT S-SHAPED BAG DELIMITER
    'lrcorner': '\u231f',  # ⌟ BOTTOM RIGHT CORNER
    'rangle': '\u27e9',  # ⟩ MATHEMATICAL RIGHT ANGLE BRACKET
    'rbag': '\u27c6',  # ⟆ RIGHT S-SHAPED BAG DELIMITER
    'rbrace': '}',  # } RIGHT CURLY BRACKET
    'rbrack': ']',  # ] RIGHT SQUARE BRACKET
    'rceil': '\u2309',  # ⌉ RIGHT CEILING
    'rfloor': '\u230b',  # ⌋ RIGHT FLOOR
    'rgroup': '\u27ef',  # ⟯ MATHEMATICAL RIGHT FLATTENED PARENTHESIS
    'rrbracket': '\u27e7',  # ⟧ MATHEMATICAL RIGHT WHITE SQUARE BRACKET
    'rrparenthesis': '\u2988',  # ⦈ Z NOTATION RIGHT IMAGE BRACKET
    'urcorner': '\u231d',  # ⌝ TOP RIGHT CORNER
    '}': '}',  # } RIGHT CURLY BRACKET
    }

mathfence = {
    'Vert': '\u2016',  # ‖ DOUBLE VERTICAL LINE
    'vert': '|',  # | VERTICAL LINE
    '|': '\u2016',  # ‖ DOUBLE VERTICAL LINE
    }

mathop = {
    'bigcap': '\u22c2',  # ⋂ N-ARY INTERSECTION
    'bigcup': '\u22c3',  # ⋃ N-ARY UNION
    'biginterleave': '\u2afc',  # ⫼ LARGE TRIPLE VERTICAL BAR OPERATOR
    'bigodot': '\u2a00',  # ⨀ N-ARY CIRCLED DOT OPERATOR
    'bigoplus': '\u2a01',  # ⨁ N-ARY CIRCLED PLUS OPERATOR
    'bigotimes': '\u2a02',  # ⨂ N-ARY CIRCLED TIMES OPERATOR
    'bigsqcap': '\u2a05',  # ⨅ N-ARY SQUARE INTERSECTION OPERATOR
    'bigsqcup': '\u2a06',  # ⨆ N-ARY SQUARE UNION OPERATOR
    'biguplus': '\u2a04',  # ⨄ N-ARY UNION OPERATOR WITH PLUS
    'bigvee': '\u22c1',  # ⋁ N-ARY LOGICAL OR
    'bigwedge': '\u22c0',  # ⋀ N-ARY LOGICAL AND
    'coprod': '\u2210',  # ∐ N-ARY COPRODUCT
    'fatsemi': '\u2a1f',  # ⨟ Z NOTATION SCHEMA COMPOSITION
    'fint': '\u2a0f',  # ⨏ INTEGRAL AVERAGE WITH SLASH
    'iiiint': '\u2a0c',  # ⨌ QUADRUPLE INTEGRAL OPERATOR
    'iiint': '\u222d',  # ∭ TRIPLE INTEGRAL
    'iint': '\u222c',  # ∬ DOUBLE INTEGRAL
    'int': '\u222b',  # ∫ INTEGRAL
    'intop': '\u222b',  # ∫ INTEGRAL
    'oiiint': '\u2230',  # ∰ VOLUME INTEGRAL
    'oiint': '\u222f',  # ∯ SURFACE INTEGRAL
    'oint': '\u222e',  # ∮ CONTOUR INTEGRAL
    'ointctrclockwise': '\u2233',  # ∳ ANTICLOCKWISE CONTOUR INTEGRAL
    'ointop': '\u222e',  # ∮ CONTOUR INTEGRAL
    'prod': '\u220f',  # ∏ N-ARY PRODUCT
    'sqint': '\u2a16',  # ⨖ QUATERNION INTEGRAL OPERATOR
    'sum': '\u2211',  # ∑ N-ARY SUMMATION
    'varointclockwise': '\u2232',  # ∲ CLOCKWISE CONTOUR INTEGRAL
    'varprod': '\u2a09',  # ⨉ N-ARY TIMES OPERATOR
    }

mathopen = {
    'Lbag': '\u27c5',  # ⟅ LEFT S-SHAPED BAG DELIMITER
    'langle': '\u27e8',  # ⟨ MATHEMATICAL LEFT ANGLE BRACKET
    'lbag': '\u27c5',  # ⟅ LEFT S-SHAPED BAG DELIMITER
    'lbrace': '{',  # { LEFT CURLY BRACKET
    'lbrack': '[',  # [ LEFT SQUARE BRACKET
    'lceil': '\u2308',  # ⌈ LEFT CEILING
    'lfloor': '\u230a',  # ⌊ LEFT FLOOR
    'lgroup': '\u27ee',  # ⟮ MATHEMATICAL LEFT FLATTENED PARENTHESIS
    'llbracket': '\u27e6',  # ⟦ MATHEMATICAL LEFT WHITE SQUARE BRACKET
    'llcorner': '\u231e',  # ⌞ BOTTOM LEFT CORNER
    'llparenthesis': '\u2987',  # ⦇ Z NOTATION LEFT IMAGE BRACKET
    'ulcorner': '\u231c',  # ⌜ TOP LEFT CORNER
    '{': '{',  # { LEFT CURLY BRACKET
    }

mathord = {
    '#': '#',  # # NUMBER SIGN
    '$': '$',  # $ DOLLAR SIGN
    '%': '%',  # % PERCENT SIGN
    '&': '&',  # & AMPERSAND
    'AC': '\u223f',  # ∿ SINE WAVE
    'APLcomment': '\u235d',  # ⍝ APL FUNCTIONAL SYMBOL UP SHOE JOT
    'APLdownarrowbox': '\u2357',  # ⍗ APL FUNCTIONAL SYMBOL QUAD DOWNWARDS ARROW
    'APLinput': '\u235e',  # ⍞ APL FUNCTIONAL SYMBOL QUOTE QUAD
    'APLinv': '\u2339',  # ⌹ APL FUNCTIONAL SYMBOL QUAD DIVIDE
    'APLleftarrowbox': '\u2347',  # ⍇ APL FUNCTIONAL SYMBOL QUAD LEFTWARDS ARROW
    'APLlog': '\u235f',  # ⍟ APL FUNCTIONAL SYMBOL CIRCLE STAR
    'APLrightarrowbox': '\u2348',  # ⍈ APL FUNCTIONAL SYMBOL QUAD RIGHTWARDS ARROW
    'APLuparrowbox': '\u2350',  # ⍐ APL FUNCTIONAL SYMBOL QUAD UPWARDS ARROW
    'Aries': '\u2648',  # ♈ ARIES
    'Box': '\u2b1c',  # ⬜ WHITE LARGE SQUARE
    'CIRCLE': '\u25cf',  # ● BLACK CIRCLE
    'CheckedBox': '\u2611',  # ☑ BALLOT BOX WITH CHECK
    'Diamond': '\u25c7',  # ◇ WHITE DIAMOND
    'Diamondblack': '\u25c6',  # ◆ BLACK DIAMOND
    'Diamonddot': '\u27d0',  # ⟐ WHITE DIAMOND WITH CENTRED DOT
    'Finv': '\u2132',  # Ⅎ TURNED CAPITAL F
    'Game': '\u2141',  # ⅁ TURNED SANS-SERIF CAPITAL G
    'Gemini': '\u264a',  # ♊ GEMINI
    'Jupiter': '\u2643',  # ♃ JUPITER
    'LEFTCIRCLE': '\u25d6',  # ◖ LEFT HALF BLACK CIRCLE
    'LEFTcircle': '\u25d0',  # ◐ CIRCLE WITH LEFT HALF BLACK
    'Leo': '\u264c',  # ♌ LEO
    'Libra': '\u264e',  # ♎ LIBRA
    'Mars': '\u2642',  # ♂ MALE SIGN
    'Mercury': '\u263f',  # ☿ MERCURY
    'Neptune': '\u2646',  # ♆ NEPTUNE
    'P': '\xb6',  # ¶ PILCROW SIGN
    'Pluto': '\u2647',  # ♇ PLUTO
    'RIGHTCIRCLE': '\u25d7',  # ◗ RIGHT HALF BLACK CIRCLE
    'RIGHTcircle': '\u25d1',  # ◑ CIRCLE WITH RIGHT HALF BLACK
    'S': '\xa7',  # § SECTION SIGN
    'Saturn': '\u2644',  # ♄ SATURN
    'Scorpio': '\u264f',  # ♏ SCORPIUS
    'Square': '\u2610',  # ☐ BALLOT BOX
    'Sun': '\u2609',  # ☉ SUN
    'Taurus': '\u2649',  # ♉ TAURUS
    'Uranus': '\u2645',  # ♅ URANUS
    'Venus': '\u2640',  # ♀ FEMALE SIGN
    'XBox': '\u2612',  # ☒ BALLOT BOX WITH X
    'Yup': '\u2144',  # ⅄ TURNED SANS-SERIF CAPITAL Y
    '_': '_',  # _ LOW LINE
    'angle': '\u2220',  # ∠ ANGLE
    'aquarius': '\u2652',  # ♒ AQUARIUS
    'aries': '\u2648',  # ♈ ARIES
    'arrowvert': '\u23d0',  # ⏐ VERTICAL LINE EXTENSION
    'backprime': '\u2035',  # ‵ REVERSED PRIME
    'backslash': '\\',  # \ REVERSE SOLIDUS
    'bigstar': '\u2605',  # ★ BLACK STAR
    'blacksmiley': '\u263b',  # ☻ BLACK SMILING FACE
    'blacksquare': '\u25fc',  # ◼ BLACK MEDIUM SQUARE
    'blacktriangle': '\u25b4',  # ▴ BLACK UP-POINTING SMALL TRIANGLE
    'blacktriangledown': '\u25be',  # ▾ BLACK DOWN-POINTING SMALL TRIANGLE
    'blacktriangleup': '\u25b4',  # ▴ BLACK UP-POINTING SMALL TRIANGLE
    'bot': '\u22a5',  # ⊥ UP TACK
    'boy': '\u2642',  # ♂ MALE SIGN
    'bracevert': '\u23aa',  # ⎪ CURLY BRACKET EXTENSION
    'cancer': '\u264b',  # ♋ CANCER
    'capricornus': '\u2651',  # ♑ CAPRICORN
    'cdots': '\u22ef',  # ⋯ MIDLINE HORIZONTAL ELLIPSIS
    'cent': '\xa2',  # ¢ CENT SIGN
    'checkmark': '\u2713',  # ✓ CHECK MARK
    'circledR': '\u24c7',  # Ⓡ CIRCLED LATIN CAPITAL LETTER R
    'circledS': '\u24c8',  # Ⓢ CIRCLED LATIN CAPITAL LETTER S
    'clubsuit': '\u2663',  # ♣ BLACK CLUB SUIT
    'complement': '\u2201',  # ∁ COMPLEMENT
    'diagdown': '\u27cd',  # ⟍ MATHEMATICAL FALLING DIAGONAL
    'diagup': '\u27cb',  # ⟋ MATHEMATICAL RISING DIAGONAL
    'diameter': '\u2300',  # ⌀ DIAMETER SIGN
    'diamondsuit': '\u2662',  # ♢ WHITE DIAMOND SUIT
    'earth': '\u2641',  # ♁ EARTH
    'emptyset': '\u2205',  # ∅ EMPTY SET
    'exists': '\u2203',  # ∃ THERE EXISTS
    'female': '\u2640',  # ♀ FEMALE SIGN
    'flat': '\u266d',  # ♭ MUSIC FLAT SIGN
    'forall': '\u2200',  # ∀ FOR ALL
    'fourth': '\u2057',  # ⁗ QUADRUPLE PRIME
    'frownie': '\u2639',  # ☹ WHITE FROWNING FACE
    'gemini': '\u264a',  # ♊ GEMINI
    'girl': '\u2640',  # ♀ FEMALE SIGN
    'heartsuit': '\u2661',  # ♡ WHITE HEART SUIT
    'infty': '\u221e',  # ∞ INFINITY
    'invdiameter': '\u2349',  # ⍉ APL FUNCTIONAL SYMBOL CIRCLE BACKSLASH
    'invneg': '\u2310',  # ⌐ REVERSED NOT SIGN
    'jupiter': '\u2643',  # ♃ JUPITER
    'ldots': '\u2026',  # … HORIZONTAL ELLIPSIS
    'leftmoon': '\u263e',  # ☾ LAST QUARTER MOON
    'leo': '\u264c',  # ♌ LEO
    'libra': '\u264e',  # ♎ LIBRA
    'lmoustache': '\u23b0',  # ⎰ UPPER LEFT OR LOWER RIGHT CURLY BRACKET SECTION
    'lnot': '\xac',  # ¬ NOT SIGN
    'lozenge': '\u25ca',  # ◊ LOZENGE
    'male': '\u2642',  # ♂ MALE SIGN
    'maltese': '\u2720',  # ✠ MALTESE CROSS
    'mathcent': '\xa2',  # ¢ CENT SIGN
    'mathdollar': '$',  # $ DOLLAR SIGN
    'mathsterling': '\xa3',  # £ POUND SIGN
    'measuredangle': '\u2221',  # ∡ MEASURED ANGLE
    'medbullet': '\u26ab',  # ⚫ MEDIUM BLACK CIRCLE
    'medcirc': '\u26aa',  # ⚪ MEDIUM WHITE CIRCLE
    'mercury': '\u263f',  # ☿ MERCURY
    'mho': '\u2127',  # ℧ INVERTED OHM SIGN
    'nabla': '\u2207',  # ∇ NABLA
    'natural': '\u266e',  # ♮ MUSIC NATURAL SIGN
    'neg': '\xac',  # ¬ NOT SIGN
    'neptune': '\u2646',  # ♆ NEPTUNE
    'nexists': '\u2204',  # ∄ THERE DOES NOT EXIST
    'notbackslash': '\u2340',  # ⍀ APL FUNCTIONAL SYMBOL BACKSLASH BAR
    'partial': '\u2202',  # ∂ PARTIAL DIFFERENTIAL
    'pisces': '\u2653',  # ♓ PISCES
    'pluto': '\u2647',  # ♇ PLUTO
    'pounds': '\xa3',  # £ POUND SIGN
    'prime': '\u2032',  # ′ PRIME
    'quarternote': '\u2669',  # ♩ QUARTER NOTE
    'rightmoon': '\u263d',  # ☽ FIRST QUARTER MOON
    'rmoustache': '\u23b1',  # ⎱ UPPER RIGHT OR LOWER LEFT CURLY BRACKET SECTION
    'sagittarius': '\u2650',  # ♐ SAGITTARIUS
    'saturn': '\u2644',  # ♄ SATURN
    'scorpio': '\u264f',  # ♏ SCORPIUS
    'second': '\u2033',  # ″ DOUBLE PRIME
    'sharp': '\u266f',  # ♯ MUSIC SHARP SIGN
    'smiley': '\u263a',  # ☺ WHITE SMILING FACE
    'spadesuit': '\u2660',  # ♠ BLACK SPADE SUIT
    'spddot': '\xa8',  # ¨ DIAERESIS
    'sphat': '^',  # ^ CIRCUMFLEX ACCENT
    'sphericalangle': '\u2222',  # ∢ SPHERICAL ANGLE
    'sptilde': '~',  # ~ TILDE
    'square': '\u25fb',  # ◻ WHITE MEDIUM SQUARE
    'sun': '\u263c',  # ☼ WHITE SUN WITH RAYS
    'surd': '\u221a',  # √ SQUARE ROOT
    'taurus': '\u2649',  # ♉ TAURUS
    'third': '\u2034',  # ‴ TRIPLE PRIME
    'top': '\u22a4',  # ⊤ DOWN TACK
    'twonotes': '\u266b',  # ♫ BEAMED EIGHTH NOTES
    'uranus': '\u2645',  # ♅ URANUS
    'varEarth': '\u2641',  # ♁ EARTH
    'varclubsuit': '\u2667',  # ♧ WHITE CLUB SUIT
    'vardiamondsuit': '\u2666',  # ♦ BLACK DIAMOND SUIT
    'varheartsuit': '\u2665',  # ♥ BLACK HEART SUIT
    'varspadesuit': '\u2664',  # ♤ WHITE SPADE SUIT
    'virgo': '\u264d',  # ♍ VIRGO
    'wasylozenge': '\u2311',  # ⌑ SQUARE LOZENGE
    'yen': '\xa5',  # ¥ YEN SIGN
    }

mathover = {
    'overbrace': '\u23de',  # ⏞ TOP CURLY BRACKET
    'wideparen': '\u23dc',  # ⏜ TOP PARENTHESIS
    }

mathpunct = {
    'ddots': '\u22f1',  # ⋱ DOWN RIGHT DIAGONAL ELLIPSIS
    'vdots': '\u22ee',  # ⋮ VERTICAL ELLIPSIS
    }

mathradical = {
    'sqrt[3]': '\u221b',  # ∛ CUBE ROOT
    'sqrt[4]': '\u221c',  # ∜ FOURTH ROOT
    }

mathrel = {
    'Bot': '\u2aeb',  # ⫫ DOUBLE UP TACK
    'Bumpeq': '\u224e',  # ≎ GEOMETRICALLY EQUIVALENT TO
    'Coloneqq': '\u2a74',  # ⩴ DOUBLE COLON EQUAL
    'Doteq': '\u2251',  # ≑ GEOMETRICALLY EQUAL TO
    'Downarrow': '\u21d3',  # ⇓ DOWNWARDS DOUBLE ARROW
    'Leftarrow': '\u21d0',  # ⇐ LEFTWARDS DOUBLE ARROW
    'Leftrightarrow': '\u21d4',  # ⇔ LEFT RIGHT DOUBLE ARROW
    'Lleftarrow': '\u21da',  # ⇚ LEFTWARDS TRIPLE ARROW
    'Longleftarrow': '\u27f8',  # ⟸ LONG LEFTWARDS DOUBLE ARROW
    'Longleftrightarrow': '\u27fa',  # ⟺ LONG LEFT RIGHT DOUBLE ARROW
    'Longmapsfrom': '\u27fd',  # ⟽ LONG LEFTWARDS DOUBLE ARROW FROM BAR
    'Longmapsto': '\u27fe',  # ⟾ LONG RIGHTWARDS DOUBLE ARROW FROM BAR
    'Longrightarrow': '\u27f9',  # ⟹ LONG RIGHTWARDS DOUBLE ARROW
    'Lsh': '\u21b0',  # ↰ UPWARDS ARROW WITH TIP LEFTWARDS
    'Mapsfrom': '\u2906',  # ⤆ LEFTWARDS DOUBLE ARROW FROM BAR
    'Mapsto': '\u2907',  # ⤇ RIGHTWARDS DOUBLE ARROW FROM BAR
    'Nearrow': '\u21d7',  # ⇗ NORTH EAST DOUBLE ARROW
    'Nwarrow': '\u21d6',  # ⇖ NORTH WEST DOUBLE ARROW
    'Perp': '\u2aeb',  # ⫫ DOUBLE UP TACK
    'Rightarrow': '\u21d2',  # ⇒ RIGHTWARDS DOUBLE ARROW
    'Rrightarrow': '\u21db',  # ⇛ RIGHTWARDS TRIPLE ARROW
    'Rsh': '\u21b1',  # ↱ UPWARDS ARROW WITH TIP RIGHTWARDS
    'Searrow': '\u21d8',  # ⇘ SOUTH EAST DOUBLE ARROW
    'Subset': '\u22d0',  # ⋐ DOUBLE SUBSET
    'Supset': '\u22d1',  # ⋑ DOUBLE SUPERSET
    'Swarrow': '\u21d9',  # ⇙ SOUTH WEST DOUBLE ARROW
    'Top': '\u2aea',  # ⫪ DOUBLE DOWN TACK
    'Uparrow': '\u21d1',  # ⇑ UPWARDS DOUBLE ARROW
    'Updownarrow': '\u21d5',  # ⇕ UP DOWN DOUBLE ARROW
    'VDash': '\u22ab',  # ⊫ DOUBLE VERTICAL BAR DOUBLE RIGHT TURNSTILE
    'Vdash': '\u22a9',  # ⊩ FORCES
    'Vvdash': '\u22aa',  # ⊪ TRIPLE VERTICAL BAR RIGHT TURNSTILE
    'apprge': '\u2273',  # ≳ GREATER-THAN OR EQUIVALENT TO
    'apprle': '\u2272',  # ≲ LESS-THAN OR EQUIVALENT TO
    'approx': '\u2248',  # ≈ ALMOST EQUAL TO
    'approxeq': '\u224a',  # ≊ ALMOST EQUAL OR EQUAL TO
    'asymp': '\u224d',  # ≍ EQUIVALENT TO
    'backepsilon': '\u220d',  # ∍ SMALL CONTAINS AS MEMBER
    'backsim': '\u223d',  # ∽ REVERSED TILDE
    'backsimeq': '\u22cd',  # ⋍ REVERSED TILDE EQUALS
    'barin': '\u22f6',  # ⋶ ELEMENT OF WITH OVERBAR
    'barleftharpoon': '\u296b',  # ⥫ LEFTWARDS HARPOON WITH BARB DOWN BELOW LONG DASH
    'barrightharpoon': '\u296d',  # ⥭ RIGHTWARDS HARPOON WITH BARB DOWN BELOW LONG DASH
    'because': '\u2235',  # ∵ BECAUSE
    'between': '\u226c',  # ≬ BETWEEN
    'blacktriangleleft': '\u25c2',  # ◂ BLACK LEFT-POINTING SMALL TRIANGLE
    'blacktriangleright': '\u25b8',  # ▸ BLACK RIGHT-POINTING SMALL TRIANGLE
    'bowtie': '\u22c8',  # ⋈ BOWTIE
    'bumpeq': '\u224f',  # ≏ DIFFERENCE BETWEEN
    'circeq': '\u2257',  # ≗ RING EQUAL TO
    'circlearrowleft': '\u21ba',  # ↺ ANTICLOCKWISE OPEN CIRCLE ARROW
    'circlearrowright': '\u21bb',  # ↻ CLOCKWISE OPEN CIRCLE ARROW
    'coloneq': '\u2254',  # ≔ COLON EQUALS
    'coloneqq': '\u2254',  # ≔ COLON EQUALS
    'cong': '\u2245',  # ≅ APPROXIMATELY EQUAL TO
    'corresponds': '\u2259',  # ≙ ESTIMATES
    'curlyeqprec': '\u22de',  # ⋞ EQUAL TO OR PRECEDES
    'curlyeqsucc': '\u22df',  # ⋟ EQUAL TO OR SUCCEEDS
    'curvearrowleft': '\u21b6',  # ↶ ANTICLOCKWISE TOP SEMICIRCLE ARROW
    'curvearrowright': '\u21b7',  # ↷ CLOCKWISE TOP SEMICIRCLE ARROW
    'dasharrow': '\u21e2',  # ⇢ RIGHTWARDS DASHED ARROW
    'dashleftarrow': '\u21e0',  # ⇠ LEFTWARDS DASHED ARROW
    'dashrightarrow': '\u21e2',  # ⇢ RIGHTWARDS DASHED ARROW
    'dashv': '\u22a3',  # ⊣ LEFT TACK
    'dlsh': '\u21b2',  # ↲ DOWNWARDS ARROW WITH TIP LEFTWARDS
    'doteq': '\u2250',  # ≐ APPROACHES THE LIMIT
    'doteqdot': '\u2251',  # ≑ GEOMETRICALLY EQUAL TO
    'downarrow': '\u2193',  # ↓ DOWNWARDS ARROW
    'downdownarrows': '\u21ca',  # ⇊ DOWNWARDS PAIRED ARROWS
    'downdownharpoons': '\u2965',  # ⥥ DOWNWARDS HARPOON WITH BARB LEFT BESIDE DOWNWARDS HARPOON WITH BARB RIGHT
    'downharpoonleft': '\u21c3',  # ⇃ DOWNWARDS HARPOON WITH BARB LEFTWARDS
    'downharpoonright': '\u21c2',  # ⇂ DOWNWARDS HARPOON WITH BARB RIGHTWARDS
    'downuparrows': '\u21f5',  # ⇵ DOWNWARDS ARROW LEFTWARDS OF UPWARDS ARROW
    'downupharpoons': '\u296f',  # ⥯ DOWNWARDS HARPOON WITH BARB LEFT BESIDE UPWARDS HARPOON WITH BARB RIGHT
    'drsh': '\u21b3',  # ↳ DOWNWARDS ARROW WITH TIP RIGHTWARDS
    'eqcirc': '\u2256',  # ≖ RING IN EQUAL TO
    'eqcolon': '\u2255',  # ≕ EQUALS COLON
    'eqqcolon': '\u2255',  # ≕ EQUALS COLON
    'eqsim': '\u2242',  # ≂ MINUS TILDE
    'eqslantgtr': '\u2a96',  # ⪖ SLANTED EQUAL TO OR GREATER-THAN
    'eqslantless': '\u2a95',  # ⪕ SLANTED EQUAL TO OR LESS-THAN
    'equiv': '\u2261',  # ≡ IDENTICAL TO
    'fallingdotseq': '\u2252',  # ≒ APPROXIMATELY EQUAL TO OR THE IMAGE OF
    'frown': '\u2322',  # ⌢ FROWN
    'ge': '\u2265',  # ≥ GREATER-THAN OR EQUAL TO
    'geq': '\u2265',  # ≥ GREATER-THAN OR EQUAL TO
    'geqq': '\u2267',  # ≧ GREATER-THAN OVER EQUAL TO
    'geqslant': '\u2a7e',  # ⩾ GREATER-THAN OR SLANTED EQUAL TO
    'gets': '\u2190',  # ← LEFTWARDS ARROW
    'gg': '\u226b',  # ≫ MUCH GREATER-THAN
    'ggcurly': '\u2abc',  # ⪼ DOUBLE SUCCEEDS
    'ggg': '\u22d9',  # ⋙ VERY MUCH GREATER-THAN
    'gggtr': '\u22d9',  # ⋙ VERY MUCH GREATER-THAN
    'gnapprox': '\u2a8a',  # ⪊ GREATER-THAN AND NOT APPROXIMATE
    'gneq': '\u2a88',  # ⪈ GREATER-THAN AND SINGLE-LINE NOT EQUAL TO
    'gneqq': '\u2269',  # ≩ GREATER-THAN BUT NOT EQUAL TO
    'gnsim': '\u22e7',  # ⋧ GREATER-THAN BUT NOT EQUIVALENT TO
    'gtrapprox': '\u2a86',  # ⪆ GREATER-THAN OR APPROXIMATE
    'gtreqless': '\u22db',  # ⋛ GREATER-THAN EQUAL TO OR LESS-THAN
    'gtreqqless': '\u2a8c',  # ⪌ GREATER-THAN ABOVE DOUBLE-LINE EQUAL ABOVE LESS-THAN
    'gtrless': '\u2277',  # ≷ GREATER-THAN OR LESS-THAN
    'gtrsim': '\u2273',  # ≳ GREATER-THAN OR EQUIVALENT TO
    'hash': '\u22d5',  # ⋕ EQUAL AND PARALLEL TO
    'hookleftarrow': '\u21a9',  # ↩ LEFTWARDS ARROW WITH HOOK
    'hookrightarrow': '\u21aa',  # ↪ RIGHTWARDS ARROW WITH HOOK
    'iddots': '\u22f0',  # ⋰ UP RIGHT DIAGONAL ELLIPSIS
    'impliedby': '\u27f8',  # ⟸ LONG LEFTWARDS DOUBLE ARROW
    'implies': '\u27f9',  # ⟹ LONG RIGHTWARDS DOUBLE ARROW
    'in': '\u2208',  # ∈ ELEMENT OF
    'le': '\u2264',  # ≤ LESS-THAN OR EQUAL TO
    'leadsto': '\u2933',  # ⤳ WAVE ARROW POINTING DIRECTLY RIGHT
    'leftarrow': '\u2190',  # ← LEFTWARDS ARROW
    'leftarrowtail': '\u21a2',  # ↢ LEFTWARDS ARROW WITH TAIL
    'leftarrowtriangle': '\u21fd',  # ⇽ LEFTWARDS OPEN-HEADED ARROW
    'leftbarharpoon': '\u296a',  # ⥪ LEFTWARDS HARPOON WITH BARB UP ABOVE LONG DASH
    'leftharpoondown': '\u21bd',  # ↽ LEFTWARDS HARPOON WITH BARB DOWNWARDS
    'leftharpoonup': '\u21bc',  # ↼ LEFTWARDS HARPOON WITH BARB UPWARDS
    'leftleftarrows': '\u21c7',  # ⇇ LEFTWARDS PAIRED ARROWS
    'leftleftharpoons': '\u2962',  # ⥢ LEFTWARDS HARPOON WITH BARB UP ABOVE LEFTWARDS HARPOON WITH BARB DOWN
    'leftrightarrow': '\u2194',  # ↔ LEFT RIGHT ARROW
    'leftrightarrows': '\u21c6',  # ⇆ LEFTWARDS ARROW OVER RIGHTWARDS ARROW
    'leftrightarrowtriangle': '\u21ff',  # ⇿ LEFT RIGHT OPEN-HEADED ARROW
    'leftrightharpoon': '\u294a',  # ⥊ LEFT BARB UP RIGHT BARB DOWN HARPOON
    'leftrightharpoons': '\u21cb',  # ⇋ LEFTWARDS HARPOON OVER RIGHTWARDS HARPOON
    'leftrightsquigarrow': '\u21ad',  # ↭ LEFT RIGHT WAVE ARROW
    'leftslice': '\u2aa6',  # ⪦ LESS-THAN CLOSED BY CURVE
    'leftsquigarrow': '\u21dc',  # ⇜ LEFTWARDS SQUIGGLE ARROW
    'leftturn': '\u21ba',  # ↺ ANTICLOCKWISE OPEN CIRCLE ARROW
    'leq': '\u2264',  # ≤ LESS-THAN OR EQUAL TO
    'leqq': '\u2266',  # ≦ LESS-THAN OVER EQUAL TO
    'leqslant': '\u2a7d',  # ⩽ LESS-THAN OR SLANTED EQUAL TO
    'lessapprox': '\u2a85',  # ⪅ LESS-THAN OR APPROXIMATE
    'lesseqgtr': '\u22da',  # ⋚ LESS-THAN EQUAL TO OR GREATER-THAN
    'lesseqqgtr': '\u2a8b',  # ⪋ LESS-THAN ABOVE DOUBLE-LINE EQUAL ABOVE GREATER-THAN
    'lessgtr': '\u2276',  # ≶ LESS-THAN OR GREATER-THAN
    'lesssim': '\u2272',  # ≲ LESS-THAN OR EQUIVALENT TO
    'lhd': '\u22b2',  # ⊲ NORMAL SUBGROUP OF
    'lightning': '\u21af',  # ↯ DOWNWARDS ZIGZAG ARROW
    'll': '\u226a',  # ≪ MUCH LESS-THAN
    'llcurly': '\u2abb',  # ⪻ DOUBLE PRECEDES
    'lll': '\u22d8',  # ⋘ VERY MUCH LESS-THAN
    'llless': '\u22d8',  # ⋘ VERY MUCH LESS-THAN
    'lnapprox': '\u2a89',  # ⪉ LESS-THAN AND NOT APPROXIMATE
    'lneq': '\u2a87',  # ⪇ LESS-THAN AND SINGLE-LINE NOT EQUAL TO
    'lneqq': '\u2268',  # ≨ LESS-THAN BUT NOT EQUAL TO
    'lnsim': '\u22e6',  # ⋦ LESS-THAN BUT NOT EQUIVALENT TO
    'longleftarrow': '\u27f5',  # ⟵ LONG LEFTWARDS ARROW
    'longleftrightarrow': '\u27f7',  # ⟷ LONG LEFT RIGHT ARROW
    'longmapsfrom': '\u27fb',  # ⟻ LONG LEFTWARDS ARROW FROM BAR
    'longmapsto': '\u27fc',  # ⟼ LONG RIGHTWARDS ARROW FROM BAR
    'longrightarrow': '\u27f6',  # ⟶ LONG RIGHTWARDS ARROW
    'looparrowleft': '\u21ab',  # ↫ LEFTWARDS ARROW WITH LOOP
    'looparrowright': '\u21ac',  # ↬ RIGHTWARDS ARROW WITH LOOP
    'lrtimes': '\u22c8',  # ⋈ BOWTIE
    'mapsfrom': '\u21a4',  # ↤ LEFTWARDS ARROW FROM BAR
    'mapsto': '\u21a6',  # ↦ RIGHTWARDS ARROW FROM BAR
    'mid': '\u2223',  # ∣ DIVIDES
    'models': '\u22a7',  # ⊧ MODELS
    'multimap': '\u22b8',  # ⊸ MULTIMAP
    'multimapboth': '\u29df',  # ⧟ DOUBLE-ENDED MULTIMAP
    'multimapdotbothA': '\u22b6',  # ⊶ ORIGINAL OF
    'multimapdotbothB': '\u22b7',  # ⊷ IMAGE OF
    'multimapinv': '\u27dc',  # ⟜ LEFT MULTIMAP
    'nLeftarrow': '\u21cd',  # ⇍ LEFTWARDS DOUBLE ARROW WITH STROKE
    'nLeftrightarrow': '\u21ce',  # ⇎ LEFT RIGHT DOUBLE ARROW WITH STROKE
    'nRightarrow': '\u21cf',  # ⇏ RIGHTWARDS DOUBLE ARROW WITH STROKE
    'nVDash': '\u22af',  # ⊯ NEGATED DOUBLE VERTICAL BAR DOUBLE RIGHT TURNSTILE
    'nVdash': '\u22ae',  # ⊮ DOES NOT FORCE
    'ncong': '\u2247',  # ≇ NEITHER APPROXIMATELY NOR ACTUALLY EQUAL TO
    'ne': '\u2260',  # ≠ NOT EQUAL TO
    'nearrow': '\u2197',  # ↗ NORTH EAST ARROW
    'neq': '\u2260',  # ≠ NOT EQUAL TO
    'ngeq': '\u2271',  # ≱ NEITHER GREATER-THAN NOR EQUAL TO
    'ngtr': '\u226f',  # ≯ NOT GREATER-THAN
    'ngtrless': '\u2279',  # ≹ NEITHER GREATER-THAN NOR LESS-THAN
    'ni': '\u220b',  # ∋ CONTAINS AS MEMBER
    'nleftarrow': '\u219a',  # ↚ LEFTWARDS ARROW WITH STROKE
    'nleftrightarrow': '\u21ae',  # ↮ LEFT RIGHT ARROW WITH STROKE
    'nleq': '\u2270',  # ≰ NEITHER LESS-THAN NOR EQUAL TO
    'nless': '\u226e',  # ≮ NOT LESS-THAN
    'nlessgtr': '\u2278',  # ≸ NEITHER LESS-THAN NOR GREATER-THAN
    'nmid': '\u2224',  # ∤ DOES NOT DIVIDE
    'notasymp': '\u226d',  # ≭ NOT EQUIVALENT TO
    'notin': '\u2209',  # ∉ NOT AN ELEMENT OF
    'notni': '\u220c',  # ∌ DOES NOT CONTAIN AS MEMBER
    'notowner': '\u220c',  # ∌ DOES NOT CONTAIN AS MEMBER
    'notslash': '\u233f',  # ⌿ APL FUNCTIONAL SYMBOL SLASH BAR
    'nparallel': '\u2226',  # ∦ NOT PARALLEL TO
    'nprec': '\u2280',  # ⊀ DOES NOT PRECEDE
    'npreceq': '\u22e0',  # ⋠ DOES NOT PRECEDE OR EQUAL
    'nrightarrow': '\u219b',  # ↛ RIGHTWARDS ARROW WITH STROKE
    'nsim': '\u2241',  # ≁ NOT TILDE
    'nsimeq': '\u2244',  # ≄ NOT ASYMPTOTICALLY EQUAL TO
    'nsubseteq': '\u2288',  # ⊈ NEITHER A SUBSET OF NOR EQUAL TO
    'nsucc': '\u2281',  # ⊁ DOES NOT SUCCEED
    'nsucceq': '\u22e1',  # ⋡ DOES NOT SUCCEED OR EQUAL
    'nsupseteq': '\u2289',  # ⊉ NEITHER A SUPERSET OF NOR EQUAL TO
    'ntriangleleft': '\u22ea',  # ⋪ NOT NORMAL SUBGROUP OF
    'ntrianglelefteq': '\u22ec',  # ⋬ NOT NORMAL SUBGROUP OF OR EQUAL TO
    'ntriangleright': '\u22eb',  # ⋫ DOES NOT CONTAIN AS NORMAL SUBGROUP
    'ntrianglerighteq': '\u22ed',  # ⋭ DOES NOT CONTAIN AS NORMAL SUBGROUP OR EQUAL
    'nvDash': '\u22ad',  # ⊭ NOT TRUE
    'nvdash': '\u22ac',  # ⊬ DOES NOT PROVE
    'nwarrow': '\u2196',  # ↖ NORTH WEST ARROW
    'owns': '\u220b',  # ∋ CONTAINS AS MEMBER
    'parallel': '\u2225',  # ∥ PARALLEL TO
    'perp': '\u27c2',  # ⟂ PERPENDICULAR
    'pitchfork': '\u22d4',  # ⋔ PITCHFORK
    'prec': '\u227a',  # ≺ PRECEDES
    'precapprox': '\u2ab7',  # ⪷ PRECEDES ABOVE ALMOST EQUAL TO
    'preccurlyeq': '\u227c',  # ≼ PRECEDES OR EQUAL TO
    'preceq': '\u2aaf',  # ⪯ PRECEDES ABOVE SINGLE-LINE EQUALS SIGN
    'preceqq': '\u2ab3',  # ⪳ PRECEDES ABOVE EQUALS SIGN
    'precnapprox': '\u2ab9',  # ⪹ PRECEDES ABOVE NOT ALMOST EQUAL TO
    'precneqq': '\u2ab5',  # ⪵ PRECEDES ABOVE NOT EQUAL TO
    'precnsim': '\u22e8',  # ⋨ PRECEDES BUT NOT EQUIVALENT TO
    'precsim': '\u227e',  # ≾ PRECEDES OR EQUIVALENT TO
    'propto': '\u221d',  # ∝ PROPORTIONAL TO
    'restriction': '\u21be',  # ↾ UPWARDS HARPOON WITH BARB RIGHTWARDS
    'rhd': '\u22b3',  # ⊳ CONTAINS AS NORMAL SUBGROUP
    'rightarrow': '\u2192',  # → RIGHTWARDS ARROW
    'rightarrowtail': '\u21a3',  # ↣ RIGHTWARDS ARROW WITH TAIL
    'rightarrowtriangle': '\u21fe',  # ⇾ RIGHTWARDS OPEN-HEADED ARROW
    'rightbarharpoon': '\u296c',  # ⥬ RIGHTWARDS HARPOON WITH BARB UP ABOVE LONG DASH
    'rightharpoondown': '\u21c1',  # ⇁ RIGHTWARDS HARPOON WITH BARB DOWNWARDS
    'rightharpoonup': '\u21c0',  # ⇀ RIGHTWARDS HARPOON WITH BARB UPWARDS
    'rightleftarrows': '\u21c4',  # ⇄ RIGHTWARDS ARROW OVER LEFTWARDS ARROW
    'rightleftharpoon': '\u294b',  # ⥋ LEFT BARB DOWN RIGHT BARB UP HARPOON
    'rightleftharpoons': '\u21cc',  # ⇌ RIGHTWARDS HARPOON OVER LEFTWARDS HARPOON
    'rightrightarrows': '\u21c9',  # ⇉ RIGHTWARDS PAIRED ARROWS
    'rightrightharpoons': '\u2964',  # ⥤ RIGHTWARDS HARPOON WITH BARB UP ABOVE RIGHTWARDS HARPOON WITH BARB DOWN
    'rightslice': '\u2aa7',  # ⪧ GREATER-THAN CLOSED BY CURVE
    'rightsquigarrow': '\u21dd',  # ⇝ RIGHTWARDS SQUIGGLE ARROW
    'rightturn': '\u21bb',  # ↻ CLOCKWISE OPEN CIRCLE ARROW
    'risingdotseq': '\u2253',  # ≓ IMAGE OF OR APPROXIMATELY EQUAL TO
    'searrow': '\u2198',  # ↘ SOUTH EAST ARROW
    'sim': '\u223c',  # ∼ TILDE OPERATOR
    'simeq': '\u2243',  # ≃ ASYMPTOTICALLY EQUAL TO
    'smile': '\u2323',  # ⌣ SMILE
    'sqsubset': '\u228f',  # ⊏ SQUARE IMAGE OF
    'sqsubseteq': '\u2291',  # ⊑ SQUARE IMAGE OF OR EQUAL TO
    'sqsupset': '\u2290',  # ⊐ SQUARE ORIGINAL OF
    'sqsupseteq': '\u2292',  # ⊒ SQUARE ORIGINAL OF OR EQUAL TO
    'strictfi': '\u297c',  # ⥼ LEFT FISH TAIL
    'strictif': '\u297d',  # ⥽ RIGHT FISH TAIL
    'subset': '\u2282',  # ⊂ SUBSET OF
    'subseteq': '\u2286',  # ⊆ SUBSET OF OR EQUAL TO
    'subseteqq': '\u2ac5',  # ⫅ SUBSET OF ABOVE EQUALS SIGN
    'subsetneq': '\u228a',  # ⊊ SUBSET OF WITH NOT EQUAL TO
    'subsetneqq': '\u2acb',  # ⫋ SUBSET OF ABOVE NOT EQUAL TO
    'succ': '\u227b',  # ≻ SUCCEEDS
    'succapprox': '\u2ab8',  # ⪸ SUCCEEDS ABOVE ALMOST EQUAL TO
    'succcurlyeq': '\u227d',  # ≽ SUCCEEDS OR EQUAL TO
    'succeq': '\u2ab0',  # ⪰ SUCCEEDS ABOVE SINGLE-LINE EQUALS SIGN
    'succeqq': '\u2ab4',  # ⪴ SUCCEEDS ABOVE EQUALS SIGN
    'succnapprox': '\u2aba',  # ⪺ SUCCEEDS ABOVE NOT ALMOST EQUAL TO
    'succneqq': '\u2ab6',  # ⪶ SUCCEEDS ABOVE NOT EQUAL TO
    'succnsim': '\u22e9',  # ⋩ SUCCEEDS BUT NOT EQUIVALENT TO
    'succsim': '\u227f',  # ≿ SUCCEEDS OR EQUIVALENT TO
    'supset': '\u2283',  # ⊃ SUPERSET OF
    'supseteq': '\u2287',  # ⊇ SUPERSET OF OR EQUAL TO
    'supseteqq': '\u2ac6',  # ⫆ SUPERSET OF ABOVE EQUALS SIGN
    'supsetneq': '\u228b',  # ⊋ SUPERSET OF WITH NOT EQUAL TO
    'supsetneqq': '\u2acc',  # ⫌ SUPERSET OF ABOVE NOT EQUAL TO
    'swarrow': '\u2199',  # ↙ SOUTH WEST ARROW
    'therefore': '\u2234',  # ∴ THEREFORE
    'to': '\u2192',  # → RIGHTWARDS ARROW
    'trianglelefteq': '\u22b4',  # ⊴ NORMAL SUBGROUP OF OR EQUAL TO
    'triangleq': '\u225c',  # ≜ DELTA EQUAL TO
    'trianglerighteq': '\u22b5',  # ⊵ CONTAINS AS NORMAL SUBGROUP OR EQUAL TO
    'twoheadleftarrow': '\u219e',  # ↞ LEFTWARDS TWO HEADED ARROW
    'twoheadrightarrow': '\u21a0',  # ↠ RIGHTWARDS TWO HEADED ARROW
    'uparrow': '\u2191',  # ↑ UPWARDS ARROW
    'updownarrow': '\u2195',  # ↕ UP DOWN ARROW
    'updownarrows': '\u21c5',  # ⇅ UPWARDS ARROW LEFTWARDS OF DOWNWARDS ARROW
    'updownharpoons': '\u296e',  # ⥮ UPWARDS HARPOON WITH BARB LEFT BESIDE DOWNWARDS HARPOON WITH BARB RIGHT
    'upharpoonleft': '\u21bf',  # ↿ UPWARDS HARPOON WITH BARB LEFTWARDS
    'upharpoonright': '\u21be',  # ↾ UPWARDS HARPOON WITH BARB RIGHTWARDS
    'upuparrows': '\u21c8',  # ⇈ UPWARDS PAIRED ARROWS
    'upupharpoons': '\u2963',  # ⥣ UPWARDS HARPOON WITH BARB LEFT BESIDE UPWARDS HARPOON WITH BARB RIGHT
    'vDash': '\u22a8',  # ⊨ TRUE
    'vartriangle': '\u25b5',  # ▵ WHITE UP-POINTING SMALL TRIANGLE
    'vartriangleleft': '\u22b2',  # ⊲ NORMAL SUBGROUP OF
    'vartriangleright': '\u22b3',  # ⊳ CONTAINS AS NORMAL SUBGROUP
    'vdash': '\u22a2',  # ⊢ RIGHT TACK
    'wasytherefore': '\u2234',  # ∴ THEREFORE
    }

mathunder = {
    'underbrace': '\u23df',  # ⏟ BOTTOM CURLY BRACKET
    }

space = {
    ' ': ' ',  #   SPACE
    ',': '\u2006',  #   SIX-PER-EM SPACE
    ':': '\u205f',  #   MEDIUM MATHEMATICAL SPACE
    'medspace': '\u205f',  #   MEDIUM MATHEMATICAL SPACE
    'quad': '\u2001',  #   EM QUAD
    'thinspace': '\u2006',  #   SIX-PER-EM SPACE
    }



from pprint import pprint

from .machaao_elements import Image

from .machaao_templates import Templates


class MachaaoButtonMessage:
    def __init__(self, text: str, buttons=None, quick_replies=None, ads=None, msg_credit=None):
        self.text = text
        self.quick_replies = quick_replies
        self.ads = ads
        self.credits = msg_credit
        self.buttons = buttons
        self.text_msg_obj = self._create_button_attachment_message_object()

    def __str__(self):
        print("MachaaoButtonMessage object")
        pprint(self.text_msg_obj, indent=2)
        return ""

    @property
    def message(self):
        return self.text_msg_obj

    def _create_button_attachment_message_object(self):
        button_msg_template = Templates().buttonMessage
        button_msg_template["message"]["attachment"]["payload"]["text"] = self.text

        if self.credits:
            button_msg_template["credit"] = self.credits

        if self.buttons is not None:
            buttons = list()
            for b in self.buttons:
                buttons.append(b.button)

            button_msg_template["message"]["attachment"]["payload"]["buttons"] = buttons

        if self.quick_replies is not None:
            quick_replies = list()
            for qr in self.quick_replies:
                quick_replies.append(qr.quick_reply)

            button_msg_template["message"]["quick_replies"] = quick_replies

        return button_msg_template


class MachaaoTextMessage:
    def __init__(self, text: str, quick_replies=None, ads=None, msg_credit=None):
        self.text = text
        self.quick_replies = quick_replies
        self.ads = ads
        self.credits = msg_credit
        self.text_msg_obj = self._create_text_message_object()

    def __str__(self):
        print("MachaaoTextMessage object")
        pprint(self.text_msg_obj, indent=2)
        return ""

    @property
    def message(self):
        return self.text_msg_obj

    def _create_text_message_object(self):
        text_msg_template = Templates().textMessage
        text_msg_template["message"]["text"] = self.text

        if self.credits:
            text_msg_template["credit"] = self.credits

        if self.ads:
            text_msg_template["ad"] = True

        if self.quick_replies is not None:
            quick_replies = list()
            for qr in self.quick_replies:
                quick_replies.append(qr.quick_reply)

            text_msg_template["message"]["quick_replies"] = quick_replies

        return text_msg_template


class MachaaoImageMessage:

    def __init__(self, image_object: Image, quick_replies=None, ads=None, msg_credit=None):
        self.img_obj = image_object.image
        self.text = self.img_obj["title"]
        self.image_url = self.img_obj["image_url"]
        self.quick_replies = quick_replies
        self.ads = ads
        self.credits = msg_credit
        self.img_msg_obj = self._create_image_message_object()

    def __str__(self):
        print("MachaaoImageMessage object")
        pprint(self.img_msg_obj, indent=2)
        return ""

    @property
    def message(self):
        return self.img_msg_obj

    def _create_image_message_object(self):
        img_msg_template = Templates().imageMessage
        img_msg_template["message"]["text"] = self.text
        img_msg_template["message"]["attachment"]["payload"]["url"] = self.image_url

        if self.quick_replies is not None:
            quick_replies = list()
            for qr in self.quick_replies:
                quick_replies.append(qr.quick_reply)

            img_msg_template["message"]["quick_replies"] = quick_replies

        return img_msg_template



"""
Serve Shared Static Files
=========================

.. autoclass:: SharedDataMiddleware
    :members: is_allowed

:copyright: 2007 Pallets
:license: BSD-3-Clause
"""

from __future__ import annotations

import importlib.util
import mimetypes
import os
import posixpath
import typing as t
from datetime import datetime
from datetime import timezone
from io import BytesIO
from time import time
from zlib import adler32

from ..http import http_date
from ..http import is_resource_modified
from ..security import safe_join
from ..utils import get_content_type
from ..wsgi import get_path_info
from ..wsgi import wrap_file

_TOpener = t.Callable[[], t.Tuple[t.IO[bytes], datetime, int]]
_TLoader = t.Callable[[t.Optional[str]], t.Tuple[t.Optional[str], t.Optional[_TOpener]]]

if t.TYPE_CHECKING:
    from _typeshed.wsgi import StartResponse
    from _typeshed.wsgi import WSGIApplication
    from _typeshed.wsgi import WSGIEnvironment


class SharedDataMiddleware:
    """A WSGI middleware which provides static content for development
    environments or simple server setups. Its usage is quite simple::

        import os
        from werkzeug.middleware.shared_data import SharedDataMiddleware

        app = SharedDataMiddleware(app, {
            '/shared': os.path.join(os.path.dirname(__file__), 'shared')
        })

    The contents of the folder ``./shared`` will now be available on
    ``http://example.com/shared/``.  This is pretty useful during development
    because a standalone media server is not required. Files can also be
    mounted on the root folder and still continue to use the application because
    the shared data middleware forwards all unhandled requests to the
    application, even if the requests are below one of the shared folders.

    If `pkg_resources` is available you can also tell the middleware to serve
    files from package data::

        app = SharedDataMiddleware(app, {
            '/static': ('myapplication', 'static')
        })

    This will then serve the ``static`` folder in the `myapplication`
    Python package.

    The optional `disallow` parameter can be a list of :func:`~fnmatch.fnmatch`
    rules for files that are not accessible from the web.  If `cache` is set to
    `False` no caching headers are sent.

    Currently the middleware does not support non-ASCII filenames. If the
    encoding on the file system happens to match the encoding of the URI it may
    work but this could also be by accident. We strongly suggest using ASCII
    only file names for static files.

    The middleware will guess the mimetype using the Python `mimetype`
    module.  If it's unable to figure out the charset it will fall back
    to `fallback_mimetype`.

    :param app: the application to wrap.  If you don't want to wrap an
                application you can pass it :exc:`NotFound`.
    :param exports: a list or dict of exported files and folders.
    :param disallow: a list of :func:`~fnmatch.fnmatch` rules.
    :param cache: enable or disable caching headers.
    :param cache_timeout: the cache timeout in seconds for the headers.
    :param fallback_mimetype: The fallback mimetype for unknown files.

    .. versionchanged:: 1.0
        The default ``fallback_mimetype`` is
        ``application/octet-stream``. If a filename looks like a text
        mimetype, the ``utf-8`` charset is added to it.

    .. versionadded:: 0.6
        Added ``fallback_mimetype``.

    .. versionchanged:: 0.5
        Added ``cache_timeout``.
    """

    def __init__(
        self,
        app: WSGIApplication,
        exports: (
            dict[str, str | tuple[str, str]]
            | t.Iterable[tuple[str, str | tuple[str, str]]]
        ),
        disallow: None = None,
        cache: bool = True,
        cache_timeout: int = 60 * 60 * 12,
        fallback_mimetype: str = "application/octet-stream",
    ) -> None:
        self.app = app
        self.exports: list[tuple[str, _TLoader]] = []
        self.cache = cache
        self.cache_timeout = cache_timeout

        if isinstance(exports, dict):
            exports = exports.items()

        for key, value in exports:
            if isinstance(value, tuple):
                loader = self.get_package_loader(*value)
            elif isinstance(value, str):
                if os.path.isfile(value):
                    loader = self.get_file_loader(value)
                else:
                    loader = self.get_directory_loader(value)
            else:
                raise TypeError(f"unknown def {value!r}")

            self.exports.append((key, loader))

        if disallow is not None:
            from fnmatch import fnmatch

            self.is_allowed = lambda x: not fnmatch(x, disallow)

        self.fallback_mimetype = fallback_mimetype

    def is_allowed(self, filename: str) -> bool:
        """Subclasses can override this method to disallow the access to
        certain files.  However by providing `disallow` in the constructor
        this method is overwritten.
        """
        return True

    def _opener(self, filename: str) -> _TOpener:
        return lambda: (
            open(filename, "rb"),
            datetime.fromtimestamp(os.path.getmtime(filename), tz=timezone.utc),
            int(os.path.getsize(filename)),
        )

    def get_file_loader(self, filename: str) -> _TLoader:
        return lambda x: (os.path.basename(filename), self._opener(filename))

    def get_package_loader(self, package: str, package_path: str) -> _TLoader:
        load_time = datetime.now(timezone.utc)
        spec = importlib.util.find_spec(package)
        reader = spec.loader.get_resource_reader(package)  # type: ignore[union-attr]

        def loader(
            path: str | None,
        ) -> tuple[str | None, _TOpener | None]:
            if path is None:
                return None, None

            path = safe_join(package_path, path)

            if path is None:
                return None, None

            basename = posixpath.basename(path)

            try:
                resource = reader.open_resource(path)
            except OSError:
                return None, None

            if isinstance(resource, BytesIO):
                return (
                    basename,
                    lambda: (resource, load_time, len(resource.getvalue())),
                )

            return (
                basename,
                lambda: (
                    resource,
                    datetime.fromtimestamp(
                        os.path.getmtime(resource.name), tz=timezone.utc
                    ),
                    os.path.getsize(resource.name),
                ),
            )

        return loader

    def get_directory_loader(self, directory: str) -> _TLoader:
        def loader(
            path: str | None,
        ) -> tuple[str | None, _TOpener | None]:
            if path is not None:
                path = safe_join(directory, path)

                if path is None:
                    return None, None
            else:
                path = directory

            if os.path.isfile(path):
                return os.path.basename(path), self._opener(path)

            return None, None

        return loader

    def generate_etag(self, mtime: datetime, file_size: int, real_filename: str) -> str:
        fn_str = os.fsencode(real_filename)
        timestamp = mtime.timestamp()
        checksum = adler32(fn_str) & 0xFFFFFFFF
        return f"wzsdm-{timestamp}-{file_size}-{checksum}"

    def __call__(
        self, environ: WSGIEnvironment, start_response: StartResponse
    ) -> t.Iterable[bytes]:
        path = get_path_info(environ)
        file_loader = None

        for search_path, loader in self.exports:
            if search_path == path:
                real_filename, file_loader = loader(None)

                if file_loader is not None:
                    break

            if not search_path.endswith("/"):
                search_path += "/"

            if path.startswith(search_path):
                real_filename, file_loader = loader(path[len(search_path) :])

                if file_loader is not None:
                    break

        if file_loader is None or not self.is_allowed(real_filename):  # type: ignore
            return self.app(environ, start_response)

        guessed_type = mimetypes.guess_type(real_filename)  # type: ignore
        mime_type = get_content_type(guessed_type[0] or self.fallback_mimetype, "utf-8")
        f, mtime, file_size = file_loader()

        headers = [("Date", http_date())]

        if self.cache:
            timeout = self.cache_timeout
            etag = self.generate_etag(mtime, file_size, real_filename)  # type: ignore
            headers += [
                ("Etag", f'"{etag}"'),
                ("Cache-Control", f"max-age={timeout}, public"),
            ]

            if not is_resource_modified(environ, etag, last_modified=mtime):
                f.close()
                start_response("304 Not Modified", headers)
                return []

            headers.append(("Expires", http_date(time() + timeout)))
        else:
            headers.append(("Cache-Control", "public"))

        headers.extend(
            (
                ("Content-Type", mime_type),
                ("Content-Length", str(file_size)),
                ("Last-Modified", http_date(mtime)),
            )
        )
        start_response("200 OK", headers)
        return wrap_file(environ, f)
